/*  
    =========================================================================

    targe2: a program for "T"riangulation of "AR"bitrary "GE"ometries in "2"D

    =========================================================================

                      Copyright: 1993, 1994 Petr Krysl 

       Czech Technical University in Prague, Faculty of Civil Engineering,
           Dept. Structural Mechanics, 166 29 Prague, Czech Republic,
                          e-mail: pk@aix.fsv.cvut.cz

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    =========================================================================
*/

#ifndef TARGE2TYPES_H
#define TARGE2TYPES_H

#include "targe2cf.h"		/* configuration specs */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#ifdef UNIX
#include <unistd.h>
#include <float.h>
#else
#define caddr_t void *
#endif

#include "listP.h"  /* to be able to use the C-toolbox */
#ifdef USE_ALISTS
#include "alistP.h"
#endif
#include "memoryP.h"

#include "utils.h"

#ifdef X_XFACE

#include <stdio.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xresource.h>
#include <X11/Intrinsic.h>

#include "Eview.h"
#include "Emodel.h"
#include "Egraphic.h"

#define HAVE_BOOLEAN

#endif

/* ===================================================================== */
/* Some convenience macros and global constants */

#ifndef FILENAME_MAX    /* POSIX constant */
#  define FILENAME_MAX 256
#endif

#ifdef MAXFILENAMECHAR
#  undef MAXFILENAMECHAR
#endif
#if FILENAME_MAX < 132
#   define MAXFILENAMECHAR 132
#else
#   define MAXFILENAMECHAR FILENAME_MAX /* POSIX constant */
#endif

#ifdef MAX_LINE_LENGTH
#  undef MAX_LINE_LENGTH
#endif
#define MAX_LINE_LENGTH 10*2048 /* maximum length of line in input file */

#ifdef SEPARATORS
#  undef SEPARATORS
#endif
#define SEPARATORS " ,[]{}\"'\n"   /* separators of tokens */

#define CONTINUATION_SEPARATOR "--"  /* separators of tokens */

#ifdef MAX_TOKENS
#  undef MAX_TOKENS
#endif
#define MAX_TOKENS 1000            /* maximum of tokens per line */

#ifdef COMMENT_CHAR
#  undef COMMENT_CHAR
#endif
#define COMMENT_CHAR '!'          /* comment char to be used in input file */

#define MAX_CONNECTIVITY  15	     /* maximum number of edges connected */
				     /* to a vertex */

#define MAX_VERT_PER_CURVE 1000      /* maximum number of vertices per curve */

#include "consts.h" /* problem-dependent constants  */

#ifndef HAVE_BOOLEAN
typedef int BOOLEAN;
#endif

/* ===================================================================== */
/* global options */

typedef struct {
  char        *in_file_name;
  BOOLEAN      use_logging;  /* YES or NO */
  char        *log_file_name;
  BOOLEAN      no_output;  /* YES or NO */
  int          output_format;
  char        *out_file_name;
  float        mesh_q_sigma;	 /* quality parameters */
  float        mesh_q_tau;
  float        mesh_q_omega;
  float        upper_lim_on_ideal_radius;
  float        merge_tolerance;
  float        cell_s_2_c_dim;
  float        remap_ratio;
  float        const_c_dim;
  float        mctl_func_exponent;
  BOOLEAN      silent_as_a_tomb;
  BOOLEAN      qtree;
  BOOLEAN      anisotropic_mesh;
  BOOLEAN      smooth_final_mesh;
  BOOLEAN      show_on_Z;
  float        Pischweitz;
  BOOLEAN      mctl_by_func;
  BOOLEAN      X_xface;
  BOOLEAN      stop_for_each;
  BOOLEAN      DOS_gxface;
  char        *gdriver_path;
  BOOLEAN      produce_quasi_estimate;
  float        hash_beta;
  float        magnific_factor;
  BOOLEAN      quality_enhancements;
#ifdef STRICT_MESH_SIZE_CONTROL
  BOOLEAN      strict_mesh_size_control;
#endif
} OPTS_REC; /* global  */
 
/* ===================================================================== */
 /* generator data structures */

typedef struct {
  float  x;
  float  y;
} COORDS_REC;

typedef struct _MCTLP_REC {
  COORDS_REC  coords;
  float       near_cdim;
  float       infl_radius;
#ifdef STRICT_MESH_SIZE_CONTROL
  float       pu;
#endif
} MCTLP_REC;

typedef LIST MCTLP_LIST;

/* sense of a curve or of an edge */
typedef enum { AS_DEFINED_SENSE = 1, REVERSED_SENSE = -1 } sense;

typedef unsigned long int CELL_ID; /* need rather large values in cell id */

struct _VERTEX_REC;
struct _CURVE_REC;
struct _EDGE_REC;
struct _SUBREG_REC;

typedef struct _VERTEX_REC {
  long int                        id;
  COORDS_REC                      coords;
  int                             connectivity;
  struct _EDGE_REC               *used_by_edges[MAX_CONNECTIVITY];
  struct _CURVE_REC              *parent_curve;
  float                           stored_cdim;
#ifdef ANISOTROPIC_MESH
  float                           stored_mf;
  float                           stored_ang;
#endif
}   VERTEX_REC; /* vertex record */

typedef VERTEX_REC *VERTEX_REC_P;

#ifndef USE_ALISTS
   typedef LIST VERTEX_LIST;
#else
   typedef ALIST VERTEX_LIST;
#endif

typedef struct _EDGE_REC {
  struct _VERTEX_REC        *vertex_1;
  struct _VERTEX_REC        *vertex_2;
  struct _CURVE_REC         *parent_curve;
  CELL_ID                   cell;
#ifdef X_XFACE
  GraphicObj                *g_p;
#endif
}   EDGE_REC;

typedef EDGE_REC *EDGE_REC_P;

#ifndef USE_ALISTS
   typedef LIST EDGE_LIST;
#else
   typedef ALIST EDGE_LIST;
#endif

typedef struct _CURVE_REF_REC {
  sense              curve_sense;
  struct _CURVE_REC     *curve_p;
} CURVE_REF_REC;

typedef struct _LOOP_REC {
  LIST            curve_ref_list;
}   LOOP_REC;

typedef LIST LOOPS_LIST;

typedef LIST C_REFS_LIST;

typedef LIST C_REFS_LIST_LIST;


typedef enum {
  DISCONTINUOUS_SPECIAL_CURVE,
  LINE_CURVE,
  CIRCLE_CURVE,
  CIRC_ARC_CURVE,
  DISCRETIZED_CURVE
} CURVE_TYPES;

typedef void (*interpolate_p)( float eta, struct _CURVE_REC *,
			      float *, float * );

typedef void (*print_p)( struct _CURVE_REC * );

typedef double (*length_p)( struct _CURVE_REC * );

typedef void (*adjust_p)( sense, sense, struct _CURVE_REC * );

typedef void (*extent_p)(float *xmin, float *ymin, float *xmax, float *ymax,
			 struct _CURVE_REC *);

typedef struct _TARGE2_METHODS_REC {
  interpolate_p          interpolate;
  print_p                      print;
  length_p                    length;
  adjust_p        adjust_to_gradient;
  extent_p                    extent;
} TARGE2_METHODS_REC;

typedef struct {
  TARGE2_METHODS_REC       *methods;
  CURVE_TYPES            curve_type;
  BOOLEAN                    meshed;
  struct _VERTEX_REC     *start_v_p;
  struct _VERTEX_REC       *end_v_p;
  sense         local_sense;
} CURVE_COMMON_REC;

typedef struct {
  COORDS_REC       first_point;
  COORDS_REC        last_point;
} LINE_DATA_REC;

typedef struct {
  COORDS_REC            first_point;
  COORDS_REC             last_point;
  float                        xmin;
  float                        ymin;
  float                        xmax;
  float                        ymax;
} DISCRETIZED_DATA_REC;

typedef struct {
  COORDS_REC           center_point;
  float                      radius;
  float                  init_angle;
} CIRCLE_DATA_REC;

typedef struct {
  COORDS_REC           center_point;
  COORDS_REC            first_point;
  COORDS_REC             last_point;
  float                      radius;
  float                 start_angle;
  float                   end_angle;
  float              internal_angle;
  BOOLEAN           in_opposite_dir;
} CIRC_ARC_DATA_REC;

typedef struct _CURVE_REC {
  CURVE_COMMON_REC      common;
  union {
    LINE_DATA_REC                line_data;
    DISCRETIZED_DATA_REC  discretized_data;
    CIRCLE_DATA_REC            circle_data;
    CIRC_ARC_DATA_REC        circ_arc_data;
  } type_depndt;
} CURVE_REC;

typedef LIST CURVE_LIST;

typedef struct _TRIANGLE_REC {
  struct _SUBREG_REC  *subreg;
  struct _EDGE_REC    *edge_1;
  struct _EDGE_REC    *edge_2;
  struct _EDGE_REC    *edge_3;
}   TRIANGLE_REC;

#ifndef USE_ALISTS
   typedef LIST TRIANGLE_LIST;
#else
   typedef ALIST TRIANGLE_LIST;
#endif

typedef struct _SUBREG_REC {
  int                         subreg_id;
  int                       property_id;
  int                       material_id;
  float                       thickness;
  int                        subsoil_id;
  char                           *group;
  C_REFS_LIST                  ext_loop;
  C_REFS_LIST_LIST            int_loops;
  BOOLEAN                        meshed;
}   SUBREG_REC;

typedef LIST SUBREG_LIST;

typedef struct _CELLS_REC {
  double              cell_size;	/* 1. */
  double              sgstd_cells_to_cdim;	/* 2. */
  double              sgstd_remap_ratio;	/* 3. */
  double              x_origin;	/* 4. */
  double              y_origin;	/* 5. */
  CELL_ID             cell_of_pull;	/* 6. */
  CELL_ID             neighbors[10]; /* 7. */
  CELL_ID             curr_neighbor;	/* 8. */
  LIST                **hyper_array_ptr;	/* 9. */
  long                hyper_n_rows;	/* 10. */
  long                hyper_n_cols;	/* 11. */
  long                leaf_array_dim;	/* 12. */
  long                n_rows;	/* 13. */
  long                n_cols;	/* 14. */
  LIST                iterator_list;	/* 15. */
  EDGE_REC            *curr_iter_edge_p; /* 16. */
  long                old_hyperar_elems;
  long                old_leafar_elems;
} CELLS_REC;

typedef struct _PULLED_EDGE_REC {
  EDGE_REC           *e_p;
  sense               sense;
  double              length;
  VERTEX_REC         *v1_p;
  COORDS_REC          first;
  float               cdim_at_first;
  float               mf_at_first;
  float               ang_at_first;
  VERTEX_REC         *v2_p;
  COORDS_REC          last;
  float               cdim_at_last;
  float               mf_at_last;
  float               ang_at_last;
  COORDS_REC          ideal_point;
  float               cdim_at_ideal;
  float               mf_at_ideal;
  float               ang_at_ideal;
  float               cdim_at_center;
} PULLED_EDGE_REC;

typedef struct _HASH_DATA_REC {
  double       min_mesh_size;
  double       logminms;
  double       max_mesh_size;
  int          first_fl_index;
  int          fl_total;
  double       one_over_log_beta;
} HASH_DATA_REC;
  
/* ===================================================================== */
 /* commands */

typedef void (*exec_proc_p)(void);

typedef struct {
  char              *keyword;
  int         keyword_length;
  exec_proc_p      handle_it;
} top_command;
  
/* ===================================================================== */

#endif
