/*  
    =========================================================================

    targe2: a program for "T"riangulation of "AR"bitrary "GE"ometries in "2"D

    =========================================================================

                      Copyright: 1993, 1994 Petr Krysl 

       Czech Technical University in Prague, Faculty of Civil Engineering,
           Dept. Structural Mechanics, 166 29 Prague, Czech Republic,
                          e-mail: pk@aix.fsv.cvut.cz

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    =========================================================================
*/
 
#include "targe2cf.h"
#include "ttypes.h" 
#include "tops.h" 
#include "mctlp.h" 
#include "vertex.h" 
#include "curve.h"
#include "loops.h"
#include "edges.h"
#include "utils.h"
#include "input.h"
#include "mesh.h"


#define ID_NOT_SET (NO_SPECIAL_VERTEX_ID)


extern OPTS_REC options; /* import global options */

static CURVE_LIST curve_list;

TARGE2_METHODS_REC line_class_rec = {
  interpolate_line,
  print_line,
  length_line,
  adj2grad_line,
  extent_line
};

TARGE2_METHODS_REC discretized_class_rec = {
  interpolate_discretized,
  print_discretized,
  length_discretized,
  adj2grad_discretized,
  extent_discretized
};

TARGE2_METHODS_REC circle_class_rec = {
  interpolate_circle,
  print_circle,
  length_circle,
  adj2grad_circle,
  extent_circle
};

TARGE2_METHODS_REC circ_arc_class_rec = {
  interpolate_circ_arc,
  print_circ_arc,
  length_circ_arc,
  adj2grad_circ_arc,
  extent_circ_arc
};

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

static float find_close_eta(CURVE_REC *c, float x, float y,
			    float eta_lower, float eta_upper);

int print_curve(NODE ptr, CURVE_REC *c);

static void subdivide_interval(CURVE_REC *c_p, float left_eta, float right_eta,
			       float left_x, float left_y,
			       float right_x, float right_y,
			       float left_d, float right_d);

static double curve_length = 0;

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

struct { /* The array is used as a transient storage which avoids */
	 /* expensive recomputation */
  float etavalue;
} precomputed_vals[MAX_VERT_PER_CURVE]; 

static int vert_num;

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

static void 
subdivide_interval(CURVE_REC *c_p, float left_eta, float right_eta,
                   float left_x, float left_y,
                   float right_x, float right_y,
                   float left_d, float right_d)
{
  float x, y, middle_eta, middle_d, dl, dr, dcl, dcr, cx, cy;
#ifdef ANISOTROPIC_MESH
  float x_plus_dx, y_plus_dy;
#endif

#define IN_TOL(v,t,R) (((v)<(t)*(R))&&((v)>(t)/(R))) 
#ifdef ANISOTROPIC_MESH
#   define IN_TOLERANCE(v,t) IN_TOL((v),(t), options.upper_lim_on_ideal_radius)
#else
#   define IN_TOLERANCE(v,t) IN_TOL((v),(t), options.upper_lim_on_ideal_radius)
#endif
#ifdef ANISOTROPIC_MESH
#   define INTERV_FRACT 1.5
#else
#   define INTERV_FRACT 1.5
#endif

  assert(left_eta >= 0 && left_eta <= 1);
  assert(right_eta >= 0 && right_eta <= 1);

  middle_eta = 0.5*(left_eta + right_eta);
  if (left_d < right_d) 
    middle_eta = min(left_eta + left_d/curve_length, middle_eta);
  else
    middle_eta = max(right_eta - right_d/curve_length, middle_eta);
  ((*c_p->common.methods).interpolate)(middle_eta, c_p, &x, &y);
#ifdef ANISOTROPIC_MESH
  ((*c_p->common.methods).interpolate)(middle_eta > 0.9999?
                                       0.9999*middle_eta: 1.0001*middle_eta,
                                       c_p, &x_plus_dx, &y_plus_dy);
  middle_d = ideal_edge_length(x, y, x_plus_dx, y_plus_dy, 0, 0, 0);
#else
  middle_d = get_mesh_cdim(x, y);
#endif

  dl       = line_length(x, y, left_x, left_y);
  dr       = line_length(x, y, right_x, right_y);
  ((*c_p->common.methods).interpolate)(0.5*(left_eta+right_eta),
                                       c_p, &cx, &cy);
  dcl      = line_length(cx, cy, left_x, left_y);
  dcr      = line_length(cx, cy, right_x, right_y);
  if ((dl+dr > INTERV_FRACT*middle_d) || dl > middle_d || dr > middle_d ||
      dcl > middle_d || dcr > middle_d ||
      (!options.qtree  &&
       (!IN_TOLERANCE(middle_d, left_d) ||
	!IN_TOLERANCE(middle_d, right_d))
       ) 
      ) {
    subdivide_interval(c_p, left_eta, middle_eta,
		       left_x, left_y, x, y, 
		       left_d, middle_d);
    subdivide_interval(c_p, middle_eta, right_eta,
		       x, y, right_x, right_y,
		       middle_d, right_d);
  } else {
    vert_num++;
    assert(vert_num < MAX_VERT_PER_CURVE); 
    precomputed_vals[vert_num].etavalue  = right_eta;
    return;
  }
}

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

void make_curve_list()
{
  curve_list = make_list();
}

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

CURVE_REC *add_curve_line(int id, float fx, float fy, float lx, float ly)
{
  CURVE_REC *c;
  make_and_add_to_tail_val(curve_list, c, CURVE_REC, "add_curve_line", id);
  c->common.methods                       = &line_class_rec;
  c->common.curve_type                    = LINE_CURVE;
  c->common.meshed                        = NO;
  c->common.start_v_p                     = NULL;
  c->common.end_v_p                       = NULL;
  c->common.local_sense                   = AS_DEFINED_SENSE;
  c->type_depndt.line_data.first_point.x  = fx;
  c->type_depndt.line_data.first_point.y  = fy;
  c->type_depndt.line_data.last_point.x   = lx;
  c->type_depndt.line_data.last_point.y   = ly;
  return c;
}

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

CURVE_REC *add_curve_discretized(int id,
			      float fx, float fy, float lx, float ly)
{
  CURVE_REC *c;
  make_and_add_to_tail_val(curve_list, c, CURVE_REC,
			   "add_curve_discretized", id);
  c->common.methods                            = &discretized_class_rec;
  c->common.curve_type                         = DISCRETIZED_CURVE;
  c->common.meshed                             = YES;
  c->common.start_v_p                          = NULL;
  c->common.end_v_p                            = NULL;
  c->common.local_sense                        = AS_DEFINED_SENSE;
  c->type_depndt.discretized_data.first_point.x= fx;
  c->type_depndt.discretized_data.first_point.y= fy;
  c->type_depndt.discretized_data.last_point.x = lx;
  c->type_depndt.discretized_data.last_point.y = ly;

  return c;
}

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

CURVE_REC *add_curve_circle(int id, float fx, float fy, float cx, float cy)
{
  CURVE_REC *c;
  double temp1, temp2;
  double fxcx, fycy;
  
  make_and_add_to_tail_val(curve_list, c, CURVE_REC, "add_curve_circle", id);
  c->common.methods                               = &circle_class_rec;
  c->common.curve_type                            = CIRCLE_CURVE;
  c->common.meshed                                = NO;
  c->common.start_v_p                             = NULL;
  c->common.end_v_p                               = NULL;
  c->common.local_sense                           = AS_DEFINED_SENSE;
  c->type_depndt.circle_data.center_point.x       = cx;
  c->type_depndt.circle_data.center_point.y       = cy;
  temp1 = fx - cx;
  temp1 *= temp1;
  temp2 = fy - cy;
  temp2 *= temp2;
  c->type_depndt.circle_data.radius = (float)sqrt(temp1 + temp2);
  fxcx = fx - cx;
  fycy = fy - cy;
  if (fabs(fxcx) < VERY_SMALL_NMBR)    {
    if (fycy < 0.)
      temp2 = 1.5 * PI;
    else
      temp2 = 0.5 * PI;
  }  else    {
    temp2 = atan(fabs(fycy / fxcx));
    if (fxcx > 0.) {
      if (fycy > 0.)
	temp2 = temp2;
      else
	temp2 = -temp2;
    }      else	{
      if (fycy > 0.) 
	temp2 = PI - temp2;
      else
	temp2 = PI + temp2;
    }
  }
  c->type_depndt.circle_data.init_angle = (float)temp2;
  return c;
}



 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

void interpolate_line(float eta, CURVE_REC *c, float *x, float *y)
{
  assert(eta >= 0 && eta <= 1);
  *x = (1. - eta) * c->type_depndt.line_data.first_point.x +
                eta * c->type_depndt.line_data.last_point.x;
  *y = (1. - eta) * c->type_depndt.line_data.first_point.y +
                eta * c->type_depndt.line_data.last_point.y;
}

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

void interpolate_circ_arc(float eta, CURVE_REC *c, float *x, float *y)
{
  double a, r;

  assert(eta >= 0 && eta <= 1);
  a  = eta * c->type_depndt.circ_arc_data.internal_angle +
    c->type_depndt.circ_arc_data.start_angle;
  r  = c->type_depndt.circ_arc_data.radius;
  *x = c->type_depndt.circ_arc_data.center_point.x + r * cos(a);
  *y = c->type_depndt.circ_arc_data.center_point.y + r * sin(a);
}

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

void interpolate_discretized(float eta, CURVE_REC *c, float *x, float *y)
{
  manip_error("Can't use interpolate for DISCRETIZED_CURVE");
}

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

void interpolate_circle(float eta, CURVE_REC *c, float *x, float *y)
{
  double a, r;

  assert(eta >= 0 && eta <= 1);
  a  = eta * PI_TIMES_2 + c->type_depndt.circle_data.init_angle;
  r  = c->type_depndt.circle_data.radius;
  *x = c->type_depndt.circle_data.center_point.x + r * cos(a);
  *y = c->type_depndt.circle_data.center_point.y + r * sin(a);
}


 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

int print_curve(NODE ptr, CURVE_REC *c)
{
  (*(*c->common.methods).print)(c);
  return 1;
}

#ifdef NEED_PRINT

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */
 /* list_curves */

void list_curves()
{
  for_all_list(curve_list, NULL, print_curve);
}

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

void print_line(CURVE_REC *c)
{
  printf("\nLine <%f;%f> <%f;%f>\n",
	 c->type_depndt.line_data.first_point.x,
	 c->type_depndt.line_data.first_point.y,
	 c->type_depndt.line_data.last_point.x,
	 c->type_depndt.line_data.last_point.y
	 );
}

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

void print_discretized(CURVE_REC *c)
{
  printf("\nDiscretized <%f;%f> <%f;%f>\n",
	 c->type_depndt.discretized_data.first_point.x,
	 c->type_depndt.discretized_data.first_point.y,
	 c->type_depndt.discretized_data.last_point.x,
	 c->type_depndt.discretized_data.last_point.y
	 );
}

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

void print_circle(CURVE_REC *c)
{
  printf("\nCircle <%f;%f> rad %f, alpha_0 %f\n",
	 c->type_depndt.circle_data.center_point.x,
	 c->type_depndt.circle_data.center_point.y,
	 c->type_depndt.circle_data.radius,
	 c->type_depndt.circle_data.init_angle
	 );
}

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

void print_circ_arc(CURVE_REC *c)
{
  printf("\nCircular arc center <%f;%f> first <%f;%f> last <%f;%f> \n \
              rad %f, alpha_start %f alpha_end %f \n",
	 c->type_depndt.circ_arc_data.center_point.x,
	 c->type_depndt.circ_arc_data.center_point.y,
	 c->type_depndt.circ_arc_data.first_point.x,
	 c->type_depndt.circ_arc_data.first_point.y,
	 c->type_depndt.circ_arc_data.last_point.x,
	 c->type_depndt.circ_arc_data.last_point.y,
	 c->type_depndt.circ_arc_data.radius,
	 c->type_depndt.circ_arc_data.start_angle,
	 c->type_depndt.circ_arc_data.end_angle
	 );
}

#else

void print_line(CURVE_REC *c)
{
}

void print_discretized(CURVE_REC *c)
{
}

void print_circle(CURVE_REC *c)
{
}

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

void print_circ_arc(CURVE_REC *c)
{
}

#endif

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

CURVE_REC  *find_curve_by_id(int curve_id)
{
  assert(curve_id > 0);
  return (CURVE_REC *)get_list_val(curve_list, curve_id);
}


 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

int get_curve_id(CURVE_REC *c)
{
  assert(c != NULL);
  return get_node_val(curve_list, c);
}

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

int mesh_curve(void *ptr, void *caddr)
{
  CURVE_REC *c_p;
  CURVE_REF_REC *c_r;
  sense s;

  c_r                     = (CURVE_REF_REC *)caddr;
  assert(c_r != NULL);
  s                       = (int)(c_r->curve_sense);
  c_p                     = c_r->curve_p;
  assert(c_p != NULL);
  c_p->common.local_sense = s;
  (*(*c_p->common.methods).adjust_to_gradient)(s, s, c_p);
  
  if (c_p->common.meshed == YES)
    add_edges_of_curve_to_front(c_p);
  else
    gen_edges_on_curve(c_p);
  
  (*(*c_p->common.methods).adjust_to_gradient)(s, c_p->common.local_sense,
					       c_p);
  return 1;
}

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

void adj2grad_line(sense s, sense local_s, CURVE_REC *c)
{
  float temp;

  assert(c != NULL);
  if (s == local_s) { /* adjust to gradient */
    if (get_mesh_cdim(c->type_depndt.line_data.first_point.x,
		      c->type_depndt.line_data.first_point.y) >
	get_mesh_cdim(c->type_depndt.line_data.last_point.x,
		      c->type_depndt.line_data.last_point.y))	{
      c->common.local_sense = (sense)(-local_s);
      SWAP_POINTS(temp, c->type_depndt.line_data.first_point.x,
		  c->type_depndt.line_data.last_point.x);
      SWAP_POINTS(temp, c->type_depndt.line_data.first_point.y,
		  c->type_depndt.line_data.last_point.y);
    }
  }  else    { /* revert */
    c->common.local_sense = (sense)(-local_s);
    SWAP_POINTS(temp, c->type_depndt.line_data.first_point.x,
		c->type_depndt.line_data.last_point.x);
    SWAP_POINTS(temp, c->type_depndt.line_data.first_point.y,
		c->type_depndt.line_data.last_point.y);
  }
}


 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

void adj2grad_discretized(sense s, sense local_s, CURVE_REC *c)
{
  ; /* Nothing to do  */
}

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

void adj2grad_circle(sense s, sense local_s, CURVE_REC *c)
{
/* do nothing for the circle */
}


 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

double length_line(CURVE_REC *c)
{
  double temp1, temp2;
  
  temp1 = (double)(c->type_depndt.line_data.last_point.x -
		   c->type_depndt.line_data.first_point.x);
  temp1 *= temp1;
  temp2 = (double)(c->type_depndt.line_data.last_point.y -
		   c->type_depndt.line_data.first_point.y);
  temp2 *= temp2;
  return sqrt(temp1 + temp2);
}

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

double 
length_discretized(CURVE_REC *c)
{
  manip_error("Can't use length for DISCRETIZED_CURVE");
  return 0.0;
}

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

double 
length_circle(CURVE_REC *c)
{
  return   c->type_depndt.circle_data.radius * PI_TIMES_2;
}

#ifdef ANISOTROPIC_MESH
#   define MF(anX, aY)  (get_mesh_2nd_dir_magn_factor(anX, aY))
#   define ANG(anX, aY) (get_mesh_1st_dir_angle(anX, aY))
#else
#   define MF(anX, aY)  (1)
#   define ANG(anX, aY) (0)
#endif


 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

void 
gen_edges_on_curve(CURVE_REC *c_p)
{ 
  float fx, fy, px, py, lx, ly;
  double length;
  int ivert, nvert, num_vertices_on_curve;
  float cdim, eta, x, y, curr_eta = 0, closest_e, lower_e, upper_e;
  VERTEX_REC *v1_p, *v2_p;
  EDGE_REC *e_p;

  assert(c_p != NULL);
  length = ((*c_p->common.methods).length)(c_p);
  num_vertices_on_curve = fixed_vertices_on_curve(c_p);
  
  ((*c_p->common.methods).interpolate)(eta = 0., c_p, &fx, &fy);
  ((*c_p->common.methods).interpolate)(eta = 1., c_p, &lx, &ly);
  
  if ((v1_p = vertex_near(fx, fy)) == NULL )
    v1_p = add_vertex(ID_NOT_SET, fx, fy, NO_CDIM_AVAILABLE, 1, 0);
  c_p->common.start_v_p = v1_p;
  set_vertex_curve_p(v1_p, c_p);

  px = fx, py = fy;
  precomputed_vals[0].etavalue  = 0.0;
  vertex_spacing(c_p, length, &nvert);
  ivert = 0;
  do {  
    ivert++;
    eta   = precomputed_vals[ivert].etavalue;
    ((*c_p->common.methods).interpolate)(eta, c_p, &x, &y);
    cdim = get_mesh_cdim(x, y);
    v2_p = NULL;
    if (num_vertices_on_curve > 0) {
      if ((v2_p = sel_fixed_on_curve(px, py, &x, &y)) != NULL) {
	num_vertices_on_curve--;
	lower_e = min(curr_eta, precomputed_vals[ivert-1].etavalue);
	upper_e = precomputed_vals[(min(nvert, (ivert+1)))].etavalue;
	closest_e = find_close_eta(c_p, x, y, lower_e, upper_e);
	curr_eta  = closest_e;
	if (ivert == nvert) 
	  ivert--;
	else
	  if (curr_eta < eta)
	    if (curr_eta < 0.333*(2*eta-precomputed_vals[ivert-1].etavalue))
	      ivert--;
      }
    } else
      curr_eta  = eta; 
    if (ivert == nvert) {
      if (v2_p == NULL) {
	x = lx; y = ly;
	if ((v2_p = vertex_near(x, y)) == NULL)
	  v2_p = add_vertex(ID_NOT_SET, x, y, cdim, MF(x, y), ANG(x, y));
      }
    } else {
      if (v2_p == NULL)
	v2_p = add_vertex(ID_NOT_SET, x, y, cdim, MF(x, y), ANG(x, y));
    }
    set_vertex_curve_p(v2_p, c_p);
    e_p = add_edge(v1_p, v2_p, c_p);
    get_vertex_coords(v2_p, &x, &y);
    add_to_front(e_p, c_p->common.local_sense, line_length(px, py, x, y));
    v1_p = v2_p;
    px   = x;
    py   = y;
  } while (ivert < nvert);

  c_p->common.end_v_p = v2_p;
  c_p->common.meshed  = YES;
}


 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

static float find_close_eta(CURVE_REC *c, float x, float y,
			    float eta_lower, float eta_upper)
{
  float step;
  float e, cx, cy, dist, d, reteta;

  step   = ((eta_upper-eta_lower)/100);
  reteta = e = eta_lower;
  ((*c->common.methods).interpolate)(e, c, &cx, &cy);
  dist = line_length(x, y, cx, cy);
  for (e = eta_lower+step; e <= eta_upper; e += step) {
    ((*c->common.methods).interpolate)(e, c, &cx, &cy);
    if ((d = line_length(x, y, cx, cy)) < dist) {
      dist   = d;
      reteta = e;
    } else 
      return reteta;
  }
  return reteta;
}

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

void vertex_spacing(CURVE_REC *c_p, double length, int *nvert)
{
  float left_eta, left_d, right_eta, right_d, lx, ly, rx, ry;
#ifdef ANISOTROPIC_MESH
  float x_plus_dx, y_plus_dy;
#endif

  assert(c_p != NULL);
  ((*c_p->common.methods).interpolate)(left_eta = 0.0, c_p, &lx, &ly);
#ifdef ANISOTROPIC_MESH
  ((*c_p->common.methods).interpolate)(0.0001, c_p, &x_plus_dx, &y_plus_dy);
  left_d    = ideal_edge_length(lx, ly, x_plus_dx, y_plus_dy, 0, 0, 0);
#else
  left_d    = get_mesh_cdim(lx, ly);
#endif
  ((*c_p->common.methods).interpolate)(right_eta = 1.0, c_p, &rx, &ry);
#ifdef ANISOTROPIC_MESH
  ((*c_p->common.methods).interpolate)(0.9999, c_p, &x_plus_dx, &y_plus_dy);
  right_d = ideal_edge_length(rx, ry, x_plus_dx, y_plus_dy, 0, 0, 0);
#else
  right_d   = get_mesh_cdim(rx, ry);
#endif

  vert_num = 0;
  precomputed_vals[vert_num].etavalue  = 0.0;
  curve_length = length;
  subdivide_interval(c_p, left_eta, right_eta,
		     lx, ly, rx, ry, left_d, right_d);
  *nvert = vert_num;
}


 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

float predict_cdim(float d1, float eta, CURVE_REC *c_p, double len1)
{
  float x, y;

  assert(c_p != NULL);
  ((*c_p->common.methods).interpolate)(eta + d1 * len1 * 0.5, c_p, &x, &y);
  return get_mesh_cdim(x, y);
}

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

void update_by_curve_extent(CURVE_REC *c_p,
		      float *xmin, float *ymin, float *xmax, float *ymax)
{
  float x_min, y_min, x_max, y_max;

  assert(c_p != NULL);
  ((*c_p->common.methods).extent)(&x_min, &y_min, &x_max, &y_max, c_p);
  *xmin = min(*xmin, x_min);
  *ymin = min(*ymin, y_min);
  *xmax = max(*xmax, x_max);
  *ymax = max(*ymax, y_max);
}

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */ 

CURVE_REC *add_curve_circ_arc(int id, 
	       float fx, float fy, /* first point */
	       float lx, float ly, /* last point */
	       BOOLEAN intermed_p_is_valid,	/* Is the following valid? */
	       float ix, float iy, /* intermediate point */
	       float cx, float cy, /* center */
	       BOOLEAN opposite_dir)
{
  CURVE_REC *c;
  double temp1, temp2, vecx, vecy, vecl, proj1, proj2, ia;
  double true_cx, true_cy, true_radius, true_start_ang, true_end_ang;
  
RAW /* Note: the routine is NOT ABLE to define half-circle correctly. */
  
  make_and_add_to_tail_val(curve_list, c, CURVE_REC, "add_curve_circ_arc", id);
  c->common.methods                               = &circ_arc_class_rec;
  c->common.curve_type                            = CIRC_ARC_CURVE;
  c->common.meshed                                = NO;
  c->common.start_v_p                             = NULL;
  c->common.end_v_p                               = NULL;
  c->common.local_sense                           = AS_DEFINED_SENSE;
  c->type_depndt.circ_arc_data.first_point.x      = fx;
  c->type_depndt.circ_arc_data.first_point.y      = fy;
  c->type_depndt.circ_arc_data.last_point.x       = lx;
  c->type_depndt.circ_arc_data.last_point.y       = ly;
  true_cx = cx; true_cy = cy;
  temp1 = fx-cx;  temp1 *= temp1;
  temp2 = fy-cy;  temp2 *= temp2;
  true_radius = sqrt(temp1 + temp2);
#define DOT_P_2(v1x, v1y, v2x, v2y) (((v1x)*(v2x))+((v1y)*(v2y)))
#define ACCURACY_EPS 1.e-7
  vecx  = lx - fx; vecy = ly - fy;
  temp1 = vecx;  temp1 *= temp1;
  temp2 = vecy;  temp2 *= temp2;
  vecl  = sqrt(temp1 + temp2);
  vecx /= vecl; vecy /= vecl;	/* unit vector */
  proj1 = DOT_P_2(cx-fx, cy-fy, vecx, vecy);
  proj2 = DOT_P_2(cx-lx, cy-ly, -vecx, -vecy);
  if (fabs(proj1-proj2) > ACCURACY_EPS) {
    true_cx = cx - 0.5 * (proj1-proj2)*vecx;
    true_cy = cy - 0.5 * (proj1-proj2)*vecy;
    temp1 = fx-true_cx;  temp1 *= temp1;
    temp2 = fy-true_cy;  temp2 *= temp2;
    true_radius = sqrt(temp1 + temp2);
  }
  vecx = fx - true_cx;
  vecy = fy - true_cy;
  if (fabs(vecx) < VERY_SMALL_NMBR)    {
    if (vecy < 0.)
      true_start_ang = 1.5 * PI;
    else
      true_start_ang = 0.5 * PI;
  }  else    {
    true_start_ang = atan(fabs(vecy / vecx));
    if (vecx > 0.) {
      if (vecy > 0.)
	true_start_ang = true_start_ang;
      else
	true_start_ang = -true_start_ang;
    }      else	{
      if (vecy > 0.) 
	true_start_ang = PI - true_start_ang;
      else
	true_start_ang = PI + true_start_ang;
    }
  }
  if (true_start_ang < 0)
    true_start_ang += 2. * PI;
  vecx = lx - true_cx;
  vecy = ly - true_cy;
  if (fabs(vecx) < VERY_SMALL_NMBR)    {
    if (vecy < 0.)
      true_end_ang = 1.5 * PI;
    else
      true_end_ang = 0.5 * PI;
  }  else    {
    true_end_ang = atan(fabs(vecy / vecx));
    if (vecx > 0.) {
      if (vecy > 0.)
	true_end_ang = true_end_ang;
      else
	true_end_ang = -true_end_ang;
    }      else	{
      if (vecy > 0.) 
	true_end_ang = PI - true_end_ang;
      else
	true_end_ang = PI + true_end_ang;
    }
  }
  if (true_end_ang < 0)
    true_end_ang += 2. * PI;

  c->type_depndt.circ_arc_data.center_point.x     = true_cx;
  c->type_depndt.circ_arc_data.center_point.y     = true_cy;
  c->type_depndt.circ_arc_data.radius             = true_radius;
  c->type_depndt.circ_arc_data.start_angle        = true_start_ang;
  c->type_depndt.circ_arc_data.end_angle          = true_end_ang;
  /*printf("%f\n",c->type_depndt.circ_arc_data.start_angle);*/
  /*printf("%f\n",c->type_depndt.circ_arc_data.end_angle);*/

  ia = true_end_ang - true_start_ang;
  if (opposite_dir) {
    if (ia > 0)
      ia = ia - 2. * PI;
    else
      ia = ia + 2. * PI;
  } else {
    if (ia>PI)
      ia = ia - 2 * PI;
    else if (ia<-PI)
      ia = ia + 2 * PI;
  }
  c->type_depndt.circ_arc_data.internal_angle = ia;
  /*printf("%f\n",c->type_depndt.circ_arc_data.internal_angle);*/

  return c;
}

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

double length_circ_arc(CURVE_REC *c)
{
  float e, ei = PI / 1000;
  float es = 0, ee = 1 + ei/10;
  float l = 0;
  float x, y, nx, ny;
  x = c->type_depndt.circ_arc_data.first_point.x;
  y = c->type_depndt.circ_arc_data.first_point.y;
  for (e = es + ei; e < ee; e += ei) {
    ((*c->common.methods).interpolate) (e, c, &nx, &ny);
    l += line_length (x, y, nx, ny);
    x = nx, y = ny;
  }
    
  return l;
}

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

void adj2grad_circ_arc(sense s, sense local_s, CURVE_REC *c)
{
  float temp;
  if (s == local_s) { /* adjust to gradient */
    if (get_mesh_cdim(c->type_depndt.circ_arc_data.first_point.x,
                      c->type_depndt.circ_arc_data.first_point.y) >
        get_mesh_cdim(c->type_depndt.circ_arc_data.last_point.x,
                      c->type_depndt.circ_arc_data.last_point.y)) {
      c->common.local_sense = (sense)(-local_s);
      SWAP_POINTS(temp,
                  c->type_depndt.circ_arc_data.first_point.x,
                  c->type_depndt.circ_arc_data.last_point.x);
      SWAP_POINTS(temp,
                  c->type_depndt.circ_arc_data.first_point.y,
                  c->type_depndt.circ_arc_data.last_point.y);
      { double a = c->type_depndt.circ_arc_data.start_angle;
        c->type_depndt.circ_arc_data.start_angle =
          c->type_depndt.circ_arc_data.end_angle;
        c->type_depndt.circ_arc_data.end_angle = a;
      }
      c->type_depndt.circ_arc_data.internal_angle
        = - c->type_depndt.circ_arc_data.internal_angle;
    }
  } else { /* revert */
    c->common.local_sense = (sense)(-local_s);
    SWAP_POINTS(temp,
                c->type_depndt.circ_arc_data.first_point.x,
                c->type_depndt.circ_arc_data.last_point.x);
    SWAP_POINTS(temp,
                c->type_depndt.circ_arc_data.first_point.y,
                c->type_depndt.circ_arc_data.last_point.y);
    { double a = c->type_depndt.circ_arc_data.start_angle;
        c->type_depndt.circ_arc_data.start_angle =
          c->type_depndt.circ_arc_data.end_angle;
        c->type_depndt.circ_arc_data.end_angle = a;
      }
    c->type_depndt.circ_arc_data.internal_angle
      = - c->type_depndt.circ_arc_data.internal_angle;
  }
}

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

void extent_line(float *xmin, float *ymin, float *xmax, float *ymax,
		 CURVE_REC *c_p)
{
  float x, y;

  *xmin = *xmax = c_p->type_depndt.line_data.first_point.x;
  *ymin = *ymax = c_p->type_depndt.line_data.first_point.y;
  x = c_p->type_depndt.line_data.last_point.x;
  y = c_p->type_depndt.line_data.last_point.y;
  *xmin = min(*xmin, x), *ymin = min(*ymin, y);
  *xmax = max(*xmax, x), *ymax = max(*ymax, y);
}

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */


void extent_circle(float *xmin, float *ymin, float *xmax, float *ymax,
		   CURVE_REC *c_p)
{
  float r, x, y;
    
  x = c_p->type_depndt.circle_data.center_point.x;
  y = c_p->type_depndt.circle_data.center_point.y;
  *xmin = *xmax = x; *ymin = *ymax = y;
  r = c_p->type_depndt.circle_data.radius;
  *xmin = min(*xmin, x+r), *ymin = min(*ymin, y);
  *xmax = max(*xmax, x+r), *ymax = max(*ymax, y);
  *xmin = min(*xmin, x), *ymin = min(*ymin, y+r);
  *xmax = max(*xmax, x), *ymax = max(*ymax, y+r);
  *xmin = min(*xmin, x-r), *ymin = min(*ymin, y);
  *xmax = max(*xmax, x-r), *ymax = max(*ymax, y);
  *xmin = min(*xmin, x), *ymin = min(*ymin, y-r);
  *xmax = max(*xmax, x), *ymax = max(*ymax, y-r);
}

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

void extent_circ_arc(float *xmin, float *ymin, float *xmax, float *ymax,
		     CURVE_REC *c_p)
{
  float e, x, y;

  ((*c_p->common.methods).interpolate)(0.0, c_p, &x, &y);
  *xmin = *xmax = x; *ymin = *ymax = y;
  for (e = 0.0; e <= 1.0; e += 0.1) {
    ((*c_p->common.methods).interpolate)(e, c_p, &x, &y);
    *xmin = min(*xmin, x), *ymin = min(*ymin, y);
    *xmax = max(*xmax, x), *ymax = max(*ymax, y);
  }
}
  
 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

void extent_discretized(float *xmin, float *ymin, float *xmax, float *ymax,
			CURVE_REC *c_p)
{
  *xmin = c_p->type_depndt.discretized_data.xmin;
  *ymin = c_p->type_depndt.discretized_data.ymin;
  *xmax = c_p->type_depndt.discretized_data.xmax;
  *ymax = c_p->type_depndt.discretized_data.ymax;
}


 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

void generate_edges_of_discretized_curve(CURVE_REC *c)
{
  int pos, tot;
  VERTEX_REC *v1_p, *v2_p;
  EDGE_REC *e_p;
  float x, y, xmin, ymin, xmax, ymax;

  assert(c != NULL);
  tot = get_tokens_total();
  pos = 4;
  x = token_float_val(pos);   y = token_float_val(pos+1);
  xmin = xmax = x; ymin = ymax = y;
  if ((v1_p = vertex_near(x, y)) == NULL) {
    v1_p = add_vertex(ID_NOT_SET, x, y, NO_CDIM_AVAILABLE, 1, 0);
    set_vertex_curve_p(v1_p, c);
  }
  c->common.start_v_p = v1_p;
  pos += 2;
  while (pos < tot) {
    x = token_float_val(pos);   y = token_float_val(pos+1);
    xmin = min(xmin, x), ymin = min(ymin, y);
    xmax = max(xmax, x), ymax = max(ymax, y);
    if (pos+2 >= tot) {
      if ((v2_p = vertex_near(x, y)) == NULL) 
	v2_p = add_vertex(ID_NOT_SET, x, y, NO_CDIM_AVAILABLE, 1, 0);
    }else {
      v2_p = add_vertex(ID_NOT_SET, x, y, NO_CDIM_AVAILABLE, 1, 0);
    }
    set_vertex_curve_p(v2_p, c);
    e_p = add_edge(v1_p, v2_p, c);
    v1_p = v2_p;
    pos += 2;
  }
  c->common.end_v_p = v2_p;
  c->type_depndt.discretized_data.xmin = xmin;
  c->type_depndt.discretized_data.ymin = ymin;
  c->type_depndt.discretized_data.xmax = xmax;
  c->type_depndt.discretized_data.ymax = ymax;
}
