/*  
    =========================================================================

    targe2: a program for "T"riangulation of "AR"bitrary "GE"ometries in "2"D

    =========================================================================

                      Copyright: 1993, 1994 Petr Krysl 

       Czech Technical University in Prague, Faculty of Civil Engineering,
           Dept. Structural Mechanics, 166 29 Prague, Czech Republic,
                          e-mail: pk@aix.fsv.cvut.cz

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    =========================================================================
*/

#include "targe2cf.h" 
#include "ttypes.h"
#include "utils.h"

#include "tops.h"
#include "vertex.h"
#include "curve.h"
#include "cells.h"
#define EDGES_PRIVATE_HDR
#include "edges.h"
#include "mesh.h"
#include "mctlp.h"

#ifdef X_XFACE

#include <stdio.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xresource.h>
#include <X11/Intrinsic.h>

#define HAVE_BOOLEAN

#include "xiface.h"

GraphicObj *draw_edge(EDGE_REC_P ep);
void hilite_pulled(void);
void unhilite_pulled(void);

static EPixel white = (EPixel)0;
static EPixel yellow = (EPixel)0;
static BOOLEAN have_white = NO;
static BOOLEAN success;


DBG(static void
draw_tentative_triangle(float fx, float fy,
                        float lx, float ly,
                        float ix, float iy);)

#endif

#ifdef DOS_GXFACE
#   define GraphicObj void
GraphicObj *draw_edge(EDGE_REC_P ep);
void hilite_pulled(void);
void unhilite_pulled(void);

#include "views.h"
#include "scr.h"
#include "status.h"



#endif

#define M_ALLOC(n) make_node(n)

EDGE_LIST edge_list;

extern OPTS_REC options; /* import global options */

extern struct _CELLS_REC cells;		/* import cells */

int          hash_by_length(double length);

HASH_DATA_REC hash; 

LIST front_lists[MAX_FRONT_LISTS];
#ifdef AF_STATISTICS
static int total_of_pulls = 0;
static int fl_counts[MAX_FRONT_LISTS];
#endif

DBG(
    int print_edge_data(NODE ptr, EDGE_REC *e_p) 
    {
      printf("edge %x\n", e_p);
    }
)


 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

void make_edge_list()
{
  edge_list = MAKE_LIST();
  assert(edge_list != NULL);
}


 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

long int total_of_edges()
{
  return LIST_LENGTH(edge_list);
}

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

EDGE_REC_P add_edge(VERTEX_REC *v1_p, VERTEX_REC *v2_p, CURVE_REC *par_c_p)
{
  static unsigned long edge_id = 0;
  EDGE_REC_P e_p;
  float fx, fy;

  edge_id++;
  MAKE_AND_ADD_TO_TAIL_VAL(edge_list, e_p, EDGE_REC, "add_edge", edge_id);
  e_p->vertex_1     = v1_p;
  e_p->vertex_2     = v2_p;
  e_p->parent_curve = par_c_p;
  get_vertex_coords(v1_p, &fx, &fy);
  e_p->cell         = get_which_cell(fx, fy);
  add_vert_reference(v1_p, e_p);
  add_vert_reference(v2_p, e_p);

#ifdef X_XFACE
  if (options.X_xface) {
    if (!have_white) {
      white = ColorGetPixelFromString("white", &success);
      yellow = ColorGetPixelFromString("yellow", &success);
    }
    e_p->g_p = draw_edge(e_p);
  } else
    e_p->g_p = NULL;
#endif
#ifdef DOS_GXFACE
  if (options.DOS_gxface) {
    draw_edge(e_p);
  }
#endif
  return e_p;
}


 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

#ifndef set_edge_cell
void set_edge_cell(EDGE_REC_P e_p, CELL_ID cell)
{
  assert(e_p != NULL);
  e_p->cell = cell;
}
#endif

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

#ifndef get_edge_cell
CELL_ID get_edge_cell(EDGE_REC_P e_p)
{
  assert(e_p != NULL);
  return e_p->cell;
}
#endif

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

void add_edges_of_curve_to_front(CURVE_REC *c_p)
{
  EDGE_REC_P e_ptr;
  float fx, fy, lx, ly;
  VERTEX_REC *v1_p = NULL, *v2_p = NULL;
  CELL_ID cell;

  assert(c_p != NULL);
  e_ptr = (EDGE_REC_P )GET_LIST_NEXT(edge_list, NULL);
  while (e_ptr != NULL) {
    if (get_edge_curve_p(e_ptr) == c_p) {
      get_edge_vertex_p(e_ptr, &v1_p, &v2_p);
      get_vertex_coords(v1_p, &fx, &fy);
      get_vertex_coords(v2_p, &lx, &ly);
      cell = get_which_cell(fx, fy);
      set_edge_cell(e_ptr, cell);
      add_to_front(e_ptr, c_p->common.local_sense,
		   line_length(fx, fy, lx, ly));
    }
    e_ptr = (EDGE_REC_P )GET_LIST_NEXT(edge_list, e_ptr);
  }
}

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

#ifndef get_edge_vertex_p
void get_edge_vertex_p(EDGE_REC_P e_p,
		       VERTEX_REC_P *v1_p, VERTEX_REC_P *v2_p )
{
  assert(e_p != NULL);
  *v1_p = e_p->vertex_1;
  *v2_p = e_p->vertex_2;
}
#endif

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

#ifndef get_edge_curve_p
CURVE_REC *get_edge_curve_p(EDGE_REC_P e_p)
{
  assert(e_p != NULL);
  return e_p->parent_curve;
}
#endif

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

void 
add_to_front(EDGE_REC_P e_p, sense s, double length)
{
  CELL_ID cell;
  int hash_ix;

  assert(e_p != NULL);
  cell                = e_p->cell;
  cells.iterator_list = get_iterator_list(cell);
  add_to_tail_val(cells.iterator_list, e_p, s);
  hash_ix = hash_by_length(length);
  if (hash_ix >= hash.fl_total)
    if (hash_ix > MAX_FRONT_LISTS)
      manip_error("Too many front list buckets");
    else
      hash.fl_total = hash_ix;
  if (front_lists[hash_ix] == NULL) 
    if ((front_lists[hash_ix] = make_list()) == NULL)
      alloc_error("add_to_front: Failed to create front_lists[hash]");
  add_to_tail_val(front_lists[hash_ix], e_p, s);
  if (hash_ix < hash.first_fl_index) 
    hash.first_fl_index = hash_ix;
#ifdef AF_STATISTICS
  total_of_pulls++;

  fl_counts[hash_ix]++;
#endif
#ifdef X_XFACE
  if (options.X_xface) {
    EMEraseGraphics(ESIModel (), e_p->g_p);
    EASValsSetColor(yellow);
    EASValsSetLineWidth(1);
    EGWithMaskChangeAttributes(WIDTH_MASK | COLOR_MASK, e_p->g_p);
    EMDrawGraphics(ESIModel (), e_p->g_p);
  }
#endif
}
 
 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */
 
int 
hash_by_length(double length)
{
  int tmp;

  assert(length > 0);
#ifdef UNIX
  tmp = (int)((log(length)-hash.logminms)*hash.one_over_log_beta) + 1;
#else
  tmp = (int)((log(length)-hash.logminms)*ONE_OVER_LOG_BETA) + 1;
#endif
  return  (tmp > 0? tmp: 0);
}
  
 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */
     
void 
init_hash_by_length(float minmsize, float maxmsize)
{
  assert(minmsize <= maxmsize && minmsize > 0);
#ifdef UNIX
  hash.one_over_log_beta    = 1./log(options.hash_beta);
#endif
  hash.min_mesh_size        = minmsize / SAFETY_MARGIN;
  hash.logminms             = log(hash.min_mesh_size);
  hash.max_mesh_size        = maxmsize * SAFETY_MARGIN;
  hash.first_fl_index       = MAX_FRONT_LISTS;
  hash.fl_total             = hash_by_length(hash.max_mesh_size);
  if (hash.fl_total > MAX_FRONT_LISTS)
    manip_error("Parameter MAX_FRONT_LISTS must be increased!!!");
}

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */
 
void reject_edge(EDGE_REC_P e_p, sense s, double length)
{
  CELL_ID cell;
  int hash_ix;

  assert(e_p != NULL);
  cell = e_p->cell;
  cells.iterator_list = get_iterator_list(cell);
  add_to_tail_val(cells.iterator_list, e_p, s);

  hash_ix = hash_by_length(length);
  add_to_tail_val(front_lists[hash_ix], e_p, s);
  if (hash_ix < hash.first_fl_index)
    hash.first_fl_index = hash_ix;
#ifdef X_XFACE
  if (options.X_xface) {
    EMEraseGraphics(ESIModel (), e_p->g_p);
    EASValsSetColor(white);
    EASValsSetLineWidth(0);
    EGWithMaskChangeAttributes(WIDTH_MASK | COLOR_MASK, e_p->g_p);
    EMDrawGraphics(ESIModel (), e_p->g_p);
  }
#endif
}

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

PULLED_EDGE_REC pulled_edge;	/* this struct keeps all information */
				/* about the pulled edge. It MUST NOT change */
				/* during the construction of the current */
				/* triangle, i.e. it MUST NOT APPEAR as */
				/* an lvalue. */

#define PE(aFIELD) pulled_edge.aFIELD

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */
 
BOOLEAN pull_edge_from_front(void)
{
  LIST fl;
  int hash_ix;
  CELL_ID cellofpull;
  float idealr;
  long row, col;
  float anchor_x, anchor_y;

  fl = front_lists[hash.first_fl_index];
  assert(fl != NULL);
  if (list_empty(fl)) {
    assert(hash.first_fl_index >=0);
    for (hash_ix = hash.first_fl_index; hash_ix <= hash.fl_total; hash_ix++) {
      if ((fl = front_lists[hash_ix]) != NULL)
	if (!list_empty(fl)) {
	  hash.first_fl_index = hash_ix;
	  goto resume;
	}
    }
#ifdef X_XFACE
    if (options.X_xface)
      unhilite_pulled();
#endif 
#ifdef DOS_GXFACE
    if (options.DOS_gxface) {
      unhilite_pulled();
    }
#endif    
#ifdef AF_STATISTICS
#   define PRINT_RATIO
    {
      int i;
      double sum = 0;

      for (i=0; i < MAX_FRONT_LISTS; i++) {
#ifdef PRINT_SUM
	if (sum != 0 && sum < 1)
	  printf("%d %f\n", i,
		 (sum += (double)fl_counts[i]/(double)total_of_pulls));
#endif
#ifdef PRINT_RATIO
	printf("%d %f\n", i, (double)fl_counts[i]/(double)total_of_pulls);
#endif
      }
    }
#endif     
    return NO;
  }  
 resume:
  PE(sense) = get_head_val(fl);
#ifdef X_XFACE
  if (options.X_xface)
    unhilite_pulled();
#endif
#ifdef DOS_GXFACE
    if (options.DOS_gxface) {
      unhilite_pulled();
    }
#endif  
  PE(e_p)   = (EDGE_REC_P)get_list_head(fl);
#ifdef X_XFACE
  if (options.X_xface)
    hilite_pulled();
#endif 
#ifdef DOS_GXFACE
    if (options.DOS_gxface) {
      hilite_pulled();
    }
#endif  
  get_edge_vertex_p(PE(e_p), &(PE(v1_p)), &(PE(v2_p)));
  if (PE(sense) == REVERSED_SENSE) {
    VERTEX_REC *p;
    
    SWAP_POINTS(p, PE(v1_p), PE(v2_p));
  }
  get_vertex_coords(PE(v1_p), &(PE(first.x)), &(PE(first.y)));
  PE(cdim_at_first) = get_stored_cdim(PE(v1_p));
#ifdef ANISOTROPIC_MESH
  PE(mf_at_first)  = get_stored_mf(PE(v1_p));
  PE(ang_at_first) = get_stored_angle(PE(v1_p));
#else
  PE(mf_at_first)  = 1;
  PE(ang_at_first) = 0;
#endif
  get_vertex_coords(PE(v2_p), &(PE(last.x)), &(PE(last.y)));
  PE(cdim_at_last) = get_stored_cdim(PE(v2_p));
#ifdef ANISOTROPIC_MESH
  PE(mf_at_last)  = get_stored_mf(PE(v2_p));
  PE(ang_at_last) = get_stored_angle(PE(v2_p));
#else
  PE(mf_at_last)  = 1;
  PE(ang_at_last) = 0;
#endif
  if (PE(sense) == REVERSED_SENSE) 
    anchor_x = PE(last.x), anchor_y = PE(last.y);
  else
    anchor_x = PE(first.x), anchor_y = PE(first.y);
  PE(length) = line_length(PE(first.x), PE(first.y), PE(last.x), PE(last.y));
  PE(cdim_at_center) = 0.5 * (PE(cdim_at_first) + PE(cdim_at_last));
  ideal_point(PE(v1_p), PE(v2_p), &(PE(ideal_point.x)), &(PE(ideal_point.y)));
#ifdef ANISOTROPIC_MESH
  PE(cdim_at_ideal) = get_mesh_cdim(PE(ideal_point.x), PE(ideal_point.y));
  PE(mf_at_ideal)   =
    get_mesh_2nd_dir_magn_factor(PE(ideal_point.x), PE(ideal_point.y));
  PE(ang_at_ideal)  =
    get_mesh_1st_dir_angle(PE(ideal_point.x), PE(ideal_point.y));
#else
  PE(cdim_at_ideal) = get_mesh_cdim(PE(ideal_point.x), PE(ideal_point.y));
  PE(mf_at_ideal)   = 1;
  PE(ang_at_ideal)  = 0;
#endif
  idealr = ideal_edge_length(0.5*(PE(first.x)+PE(last.x)),
                             0.5*(PE(first.y)+PE(last.y)),
                             PE(ideal_point.x), PE(ideal_point.y),
                             PE(cdim_at_ideal),
                             PE(mf_at_ideal),
                             PE(ang_at_ideal));
  DBG(draw_tentative_triangle(PE(first.x), PE(first.y), PE(last.x), PE(last.y),
                          PE(ideal_point.x), PE(ideal_point.y));)
  if (idealr > cells.cell_size * cells.sgstd_remap_ratio) {
    rearrange_cells(cells.cell_size, options.magnific_factor *
		    idealr * cells.sgstd_cells_to_cdim); 
    remap_to_new_cells();
  } 
  cellofpull = get_which_cell(anchor_x, anchor_y);
  assert(cellofpull >= 1 && cellofpull <= cells.n_rows*cells.n_cols);
  row        = (cellofpull - 1) / cells.n_cols + 1;
  col        = cellofpull - (row - 1) * cells.n_cols;
  cells.neighbors[CIX]   = cellofpull;
  cells.neighbors[RUPIX] = 0;
  cells.neighbors[RIX]   = 0;
  cells.neighbors[RLOIX] = 0;
  cells.neighbors[LUPIX] = 0;
  cells.neighbors[LIX]   = 0;
  cells.neighbors[LLOIX] = 0;
  cells.neighbors[LOIX]  = 0;
  cells.neighbors[UPIX]  = 0;
  if (row < cells.n_rows)
    cells.neighbors[UPIX]  = cellofpull + cells.n_cols;
  if (row > 1) 
    cells.neighbors[LOIX]  = cellofpull - cells.n_cols;
  if (col < cells.n_cols) {
    cells.neighbors[RIX] = cellofpull + 1;
    if (row > 1) 
      cells.neighbors[RLOIX] = cells.neighbors[LOIX] + 1;
    if (row < cells.n_rows) 
      cells.neighbors[RUPIX] = cells.neighbors[UPIX] + 1;
  }
  if (col > 1) {
    cells.neighbors[LIX] = cellofpull - 1;
    if (row > 1) 
      cells.neighbors[LLOIX] = cells.neighbors[LOIX] - 1;
    if (row < cells.n_rows) 
      cells.neighbors[LUPIX] = cells.neighbors[UPIX] - 1;
  }
  cells.iterator_list = get_iterator_list(cellofpull);
  assert(cells.iterator_list != NULL);
  {
    EDGE_REC_P p;
    
    p = (EDGE_REC_P)get_list_next(cells.iterator_list, NULL);
    while(p != NULL) {
      if (p == PE(e_p)) {
	if (PE(sense) == get_node_val_current(cells.iterator_list)) {
	  p = (EDGE_REC_P)get_from_current(cells.iterator_list);
	  p = (EDGE_REC_P)get_from_head(fl);
	  goto exit;
	}
      }
      p = (EDGE_REC_P)get_list_next(cells.iterator_list, p);
    }
    manip_error("pull_edge_from_front: Edge not in iterlist !?");
  }
 exit:
#ifdef X_XFACE
  if (options.X_xface) {
    EMEraseGraphics(ESIModel (), PE(e_p)->g_p);
    EASValsSetColor(white);
    EASValsSetLineWidth(0);
    EGWithMaskChangeAttributes(WIDTH_MASK | COLOR_MASK, PE(e_p)->g_p);
    EMDrawGraphics(ESIModel (), PE(e_p)->g_p);
  }
#endif  
  return YES;
}

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

LIST get_iterator_list(CELL_ID cell)
{
  long row, col;
  long int i, len;
  long hyperrow, hypercol;
  CELL_ID hypercell, leafcell;
  LIST *leafarrayptr;
  LIST ilp = NULL;

  assert(cell >= 1 && cell <= cells.n_cols*cells.n_rows);
  row = (long)(((double)cell - 1)/((double)cells.n_cols)) + 1;
  assert(row >= 1 && row <= cells.n_rows);
  col = cell - (row - 1) * cells.n_cols;
  assert(col >= 1 && col <= cells.n_cols);

  hyperrow  = (long)((row - 1)/(double)cells.leaf_array_dim) + 1;
  hypercol  = (long)((col - 1)/(double)cells.leaf_array_dim) + 1;
  hypercell = (hyperrow - 1)*cells.hyper_n_cols + hypercol; 

  if ((leafarrayptr =
       (*((LIST **)(cells.hyper_array_ptr + hypercell)))) == NULL) {
    len = cells.leaf_array_dim*cells.leaf_array_dim;
    if ((leafarrayptr = (LIST *)M_ALLOC(len*sizeof(LIST *))) == NULL) {
      alloc_error("get_iterator_list");
      return NULL;
    }
    for(i=0;i<len;i++)
      (*((LIST *)(leafarrayptr + i))) = NULL;
    (*((LIST **)(cells.hyper_array_ptr + hypercell))) = leafarrayptr;
  }

  leafcell = (row - cells.leaf_array_dim * (hyperrow - 1) - 1) *
    cells.leaf_array_dim + (col - cells.leaf_array_dim * (hypercol - 1));

  if ((ilp = (*((LIST *)(leafarrayptr + leafcell - 1)))) == NULL) {
    ilp = make_list();
    (*((LIST *)(leafarrayptr + leafcell - 1))) = ilp;
  }

  assert(ilp != NULL);
  return ilp;
}

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */
 
BOOLEAN destroy_iter_lists()
{
  long hyperelem, leafelem, k, m;
  LIST *la_ptr;
  LIST itlistptr;

  assert(cells.hyper_array_ptr != NULL);
  assert(*((long *)(cells.hyper_array_ptr)) == cells.old_hyperar_elems);
  hyperelem   = cells.old_hyperar_elems;
  leafelem    = cells.old_leafar_elems;
  for (k = 1; k <= hyperelem; k++) {
    if ((la_ptr = (*(LIST **)(cells.hyper_array_ptr + k))) != NULL) {
      for (m = 0; m < leafelem; m++) {
	if ((itlistptr = (*((LIST *)(la_ptr + m)))) != NULL) 
	  free_list(itlistptr, NOT_NODES);
      }
      free_node(la_ptr);
    }
  }
  free_node(cells.hyper_array_ptr);
  cells.hyper_array_ptr = NULL;
  return YES;
}

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

BOOLEAN create_iter_lists()
{
  long len, i;
  
  len = cells.hyper_n_rows * cells.hyper_n_cols + 1;
  if ((cells.hyper_array_ptr = (LIST **)M_ALLOC(len*sizeof(LIST *))) == NULL)
    alloc_error("create_iter_lists");
  len = cells.hyper_n_rows * cells.hyper_n_cols;
  for (i=1;i<=len;i++)
    *(cells.hyper_array_ptr+i) = NULL;
  *((long *)(cells.hyper_array_ptr)) = len;
  return YES;
}

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

void first_front_edge(EDGE_REC_P *epp, sense *s)
{
  int cn;
  LIST ilp;
  CELL_ID cell;
  EDGE_REC_P e_;
  sense tmp_s;

  e_    = NULL;
  tmp_s = 1;
  for (cn = 1; cn <= 9; cn++) {
    cell = cells.neighbors[cn];
    if (cell > 0) {
      ilp = get_iterator_list(cell);
      if (ilp != NULL) {
	if (!list_empty(ilp)) {
	  cells.curr_neighbor = cn;
	  cells.iterator_list = ilp;
	  if ((e_ = (EDGE_REC_P )get_list_next(ilp, NULL)) != NULL) 
	    tmp_s = get_node_val_current(ilp);
	  cells.curr_iter_edge_p = e_;
	  goto exit;
	}
      }
    }
  }
 exit:
  *epp = e_;
  *s   = tmp_s;
}
 
 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

void next_front_edge(EDGE_REC_P *epp, sense *s)
{
  EDGE_REC_P e_;
  CELL_ID cell;
  sense tmp_s;

  e_ = (EDGE_REC_P)get_list_next(cells.iterator_list, cells.curr_iter_edge_p);
  if (e_ == NULL) {
  retry:
    cells.curr_neighbor++;
    if (cells.curr_neighbor > 9) {
      e_    = NULL;
      tmp_s = 1;
      goto exit;
    }
    cell = cells.neighbors[cells.curr_neighbor];
    if (cell > 0) {
      cells.iterator_list = get_iterator_list(cell);
      if (cells.iterator_list != NULL) {
	if (!list_empty(cells.iterator_list)) {
	  if ((e_ = (EDGE_REC_P )get_list_next(cells.iterator_list, NULL))
	      != NULL)
	    tmp_s = get_node_val_current(cells.iterator_list);
	  cells.curr_iter_edge_p = e_;
	  goto exit;
	}
      }
    }
    goto retry;
  } else {
    cells.curr_iter_edge_p = e_;
    tmp_s = get_node_val_current(cells.iterator_list);
  }
 exit:
  *epp = e_;
  *s   = tmp_s;
}
 
 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */
 
void remove_edge_from_front(EDGE_REC_P ep, sense sns)
{
  VERTEX_REC *v1_p = NULL, *v2_p = NULL;
  int hashval;
  LIST flptr;
  EDGE_REC_P e_;
  float fx, fy, lx, ly;
  CELL_ID cell;

  assert(ep != NULL);
  get_edge_vertex_p(ep, &v1_p, &v2_p);
  get_vertex_coords(v1_p, &fx, &fy);
  get_vertex_coords(v2_p, &lx, &ly);
  hashval = hash_by_length(line_length(fx, fy, lx, ly));
  flptr   = front_lists[hashval];
  assert(flptr != NULL);
  if (list_empty(flptr))
    manip_error("remove_edge_from_front: List empty!?");
  
  e_ = (EDGE_REC_P)get_list_next(flptr, NULL);
  while (e_ != NULL) {
    if (e_ == ep && ((sense)get_node_val_current(flptr) == sns))
      goto found_in_front;
    e_ = (EDGE_REC_P )get_list_next(flptr, e_);
  }
  manip_error("remove_edge_from_front: Edge not in front list!?");

 found_in_front:
  cell = get_edge_cell(ep);
  cells.iterator_list = get_iterator_list(cell);
  e_   = (EDGE_REC_P)get_from_current(flptr);

  e_ = (EDGE_REC_P)get_list_next(cells.iterator_list, NULL);
  while (e_ != NULL) {
    if (e_ == ep &&
	((sense)get_node_val_current(cells.iterator_list) == sns)) {
      e_ = (EDGE_REC_P )get_from_current(cells.iterator_list);
      goto final;
    }
    e_ = (EDGE_REC_P )get_list_next(cells.iterator_list, e_);
  }
  manip_error("remove_edge_from_front: Edge not in iterator list!?");
 final:
#ifdef X_XFACE
  if (options.X_xface) {
    EMEraseGraphics(ESIModel (), ep->g_p);
    EASValsSetColor(white);
    EASValsSetLineWidth(0);
    EGWithMaskChangeAttributes(WIDTH_MASK | COLOR_MASK, ep->g_p);
    EMDrawGraphics(ESIModel (), ep->g_p);
  }
#endif
  ;
}

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */
 
#ifdef X_XFACE

static char *
edge_info(GraphicObj *);

static EGMethodsRec *edge_methods;

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */
 
GraphicObj *
draw_edge(EDGE_REC_P ep)
{
  float x1, x2, y1, y2;
  GraphicObj *g_p;
  WCRec pnts[2] = {0, 0, 0, 0, 0, 0};
  static int first = 1;
  
  get_vertex_coords(ep->vertex_1, &x1, &y1);
  get_vertex_coords(ep->vertex_2, &x2, &y2);

#define IPOL_LINE(eta, v1, v2) (((1. - (eta)) * (v1)) + ((eta) * (v2)))

  pnts[0].x = IPOL_LINE(0.05, x1, x2);
  pnts[0].y = IPOL_LINE(0.05, y1, y2);
  pnts[1].x = IPOL_LINE(0.95, x1, x2);
  pnts[1].y = IPOL_LINE(0.95, y1, y2);
  if (options.show_on_Z) {
    if        (options.show_on_Z == 1) {
      pnts[0].z = get_stored_cdim(ep->vertex_1);
      pnts[1].z = get_stored_cdim(ep->vertex_2);
    } else if (options.show_on_Z == 2) {
      pnts[0].z = get_stored_mf(ep->vertex_1) * Z_SHOW_SCALE;
      pnts[1].z = get_stored_mf(ep->vertex_2) * Z_SHOW_SCALE;
    } else if (options.show_on_Z == 3) {
      pnts[0].z = get_stored_angle(ep->vertex_1) * Z_SHOW_SCALE;
      pnts[1].z = get_stored_angle(ep->vertex_2) * Z_SHOW_SCALE;
    }
  }
#ifndef DONT_DRAW_TRIANGLES
#   ifdef LINE_EDGE_CODED
  g_p = CreateLineWithData(fx, fy, lx, ly,
			   1./get_stored_cdim(ep->vertex_1),
			   1./get_stored_cdim(ep->vertex_2));
  LineSetAttributes(g_p, WIDTH_MASK, 0, 0, 0);
#   else
  g_p = CreateLine3D(pnts);
  if (!have_white) {
    white = ColorGetPixelFromString("white", &success);
    yellow = ColorGetPixelFromString("yellow", &success);
  }
  EASValsSetLineWidth(0);
  EASValsSetColor(white);
  EGWithMaskChangeAttributes(WIDTH_MASK | COLOR_MASK, g_p);
#   endif
#else
#   ifdef LINE_EDGE_CODED
  g_p = CreateLineWithData(fx, fy, lx, ly,
			   1./get_stored_cdim(ep->vertex_1),
			   1./get_stored_cdim(ep->vertex_2));
  EASValsSetLineWidth(0);
  EASValsSetLineStyle(SOLID_STYLE);
  EGWithMaskChangeAttributes(WIDTH_MASK, g_p);
#   else
  g_p = CreateLine3D(pnts);
  if (!have_white) 
    white = ColorGetPixelFromString("white", &success);
  EASValsSetLineWidth(0);
  EASValsSetColor(white);
  EGWithMaskChangeAttributes(WIDTH_MASK | COLOR_MASK, g_p);
#endif
#endif
  EMAddGraphicsToModel(ESIModel (), g_p);
  if (first) {
    EMaskArray ma = {0, 0};
    EGMethodsRec mt;
    EGSetOverrideEntry(ma, (&mt), string_rep_func, edge_info);
    edge_methods = EGCopyMethodTable(g_p, ma, &mt);
    first = 0;
  }
  EGAttachObject(g_p, (EObjectP)ep);
  EGOverrideMethodTable(g_p, edge_methods);
  
  return g_p;
}
 
 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */
 
void draw_edges()
{
  EDGE_REC *ep;
  VERTEX_REC *v1 = NULL, *v2 = NULL;
  float fx, fy, lx, ly;
  WCRec pnts[2] = {0, 0, 0, 0, 0, 0};

  ep = (EDGE_REC *)GET_LIST_NEXT(edge_list, NULL);
  while (ep != NULL) {
    EMEraseGraphics(ESIModel (), ep->g_p);
    get_edge_vertex_p(ep, &v1, &v2);
    get_vertex_coords(v1, &fx, &fy);
    get_vertex_coords(v2, &lx, &ly);
    pnts[0].x = fx, pnts[0].y = fy;
    pnts[1].x = lx, pnts[1].y = ly;
    if (options.show_on_Z) {
      if        (options.show_on_Z == 1) {
        pnts[0].z = get_stored_cdim(ep->vertex_1);
        pnts[1].z = get_stored_cdim(ep->vertex_2);
      } else if (options.show_on_Z == 2) {
        pnts[0].z = get_stored_mf(ep->vertex_1) * Z_SHOW_SCALE;
        pnts[1].z = get_stored_mf(ep->vertex_2) * Z_SHOW_SCALE;
      } else if (options.show_on_Z == 3) {
        pnts[0].z = get_stored_angle(ep->vertex_1) * Z_SHOW_SCALE;
        pnts[1].z = get_stored_angle(ep->vertex_2) * Z_SHOW_SCALE;
      }
    }
    EGModifyGraphicsGeometry(ep->g_p, pnts);
    EMDrawGraphics(ESIModel (), ep->g_p);
    ep = (EDGE_REC *)GET_LIST_NEXT(edge_list, ep);
  }
}
 
 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */
 
void hilite_pulled(void)
{
  extern PULLED_EDGE_REC pulled_edge;
  
  EMHiliteGraphics(ESIModel (), (PE(e_p))->g_p);
}
 
 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */
 
void unhilite_pulled(void)
{
  extern PULLED_EDGE_REC pulled_edge;
  if (PE(e_p) != NULL)
    EMUnhiliteGraphics(ESIModel (), (PE(e_p))->g_p);
}

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */
 
static char buf[1024];

static char *
edge_info(GraphicObj *gp)
{
  EDGE_REC *ep;
  float x1, y1, x2, y2;
  VERTEX_REC *v1 = NULL, *v2 = NULL;
  static int drew_quads = 0;
  /*void draw_leaves(void);*/
  CELL_ID cell;

  if (!drew_quads) {
    /*draw_leaves();*/
    drew_quads = 1;
  }
  ep = (EDGE_REC *)EGGetAppObj(gp);
  get_edge_vertex_p(ep, &v1, &v2);
  get_vertex_coords(v1, &x1, &y1);
  get_vertex_coords(v2, &x2, &y2);
  cell = get_which_cell(x1, y1); 
  sprintf(buf,
          "Edge at 0x%x <%f;%f> <%f;%f> (length %f, cell %ld, mcdim %f, %f)",
          ep, x1, y1, x2, y2, line_length(x1, y1, x2, y2), cell,
          get_mesh_cdim(x1, y1), get_mesh_cdim(x2, y2));
  return buf;
}


DBG(
static GraphicObj *tt;
static WCRec ttpnts[3];
static EPixel ttcolor = 0x0;
static BOOLEAN ttsuc;

static void
draw_tentative_triangle(float fx, float fy,
                        float lx, float ly,
                        float ix, float iy)
{
  if (tt == NULL) {
    tt = CreateTriangle3D(ttpnts);
    ttcolor = ColorGetPixelFromString("hotpink", &ttsuc);
    EASValsSetColor(ttcolor);
    EASValsSetFillStyle(FILL_HOLLOW);
    EGWithMaskChangeAttributes(COLOR_MASK | FILL_MASK, tt);
    EMAddGraphicsToModel(ESIModel (), tt);
  }
  EMEraseGraphics(ESIModel (), tt);
  ttpnts[0].x = fx;
  ttpnts[0].y = fy;
  ttpnts[1].x = lx;
  ttpnts[1].y = ly;
  ttpnts[2].x = ix;
  ttpnts[2].y = iy;
  EGModifyGraphicsGeometry(tt, ttpnts);
  EMDrawGraphics(ESIModel (), tt);
}
)


void
print_edge_data(FILE *fp, EObjectP p)
{
    EDGE_REC *ep;
    float x, y;
    
    if (p == NULL) return;

    ep = (EDGE_REC *)p;
    
    fprintf(fp, "Edge\n");
    get_vertex_coords(ep->vertex_1, &x, &y);
    fprintf(fp, "\tVertex 1: \t\t<%g;%g>\n\t\t\t\tS=%g;M=%g;A=%g\n", x, y,
            get_stored_cdim(ep->vertex_1),
            get_stored_mf(ep->vertex_1),
            get_stored_angle(ep->vertex_1));
    get_vertex_coords(ep->vertex_2, &x, &y);
    fprintf(fp, "\tVertex 2: \t\t<%g;%g>\n\t\t\t\tS=%g;M=%g;A=%g\n", x, y,
            get_stored_cdim(ep->vertex_2),
            get_stored_mf(ep->vertex_2),
            get_stored_angle(ep->vertex_2));
}


#endif

#ifdef DOS_GXFACE
 
 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

GraphicObj *draw_edge(EDGE_REC_P ep)
{
  float x1, x2, y1, y2, fx, fy, lx, ly;
  GraphicObj *g_p;
  BOOLEAN success;

  assert(ep != NULL);
  get_vertex_coords(ep->vertex_1, &x1, &y1);
  get_vertex_coords(ep->vertex_2, &x2, &y2);

#define IPOL_LINE(eta, v1, v2) (((1. - (eta)) * (v1)) + ((eta) * (v2)))

  fx = IPOL_LINE(0.025, x1, x2);
  fy = IPOL_LINE(0.025, y1, y2);
  lx = IPOL_LINE(0.975, x1, x2);
  ly = IPOL_LINE(0.975, y1, y2);
#ifndef DONT_DRAW_TRIANGLES
/* RAW: draw line in color encoding */
  setcolor(DRAW_AREA_INK);
  lineview (fx,fy,lx,ly);
#else
/* RAW: draw in normal rendering color */
  setcolor(DRAW_AREA_INK);
  lineview (fx,fy,lx,ly);
#endif
  return (GraphicObj *)NULL;
}

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */
 
void draw_edges()
{
  EDGE_REC *ep;
  VERTEX_REC *v1, *v2;
  float fx, fy, lx, ly;

 /* RAW: clear screen */
  ep = (EDGE_REC *)GET_LIST_NEXT(edge_list, NULL);
  while (ep != NULL) {
    draw_edge(ep);
    ep = (EDGE_REC *)GET_LIST_NEXT(edge_list, ep);
  }
}

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

void hilite_pulled(void)
{
  extern PULLED_EDGE_REC pulled_edge;

/* RAW: highlight line */
}

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

void unhilite_pulled(void)
{
  extern PULLED_EDGE_REC pulled_edge;

  if (PE(e_p) != NULL)
/* RAW: unhighlight line */;
}

#endif /* DOS_GXFACE */


void       
dump_edges_in_mos_style(FILE *fp)
{
  EDGE_REC *ep;
  long eid = 0;

  /* IMPORTANT NOTICE: */
  /* The routine sets the edge id into the /cell/ field.  This value is */
  /* then used by dump_triangles_in_mos_style(); */
  /* */
  ep = (EDGE_REC *)GET_LIST_NEXT(edge_list, NULL);
  while (ep != NULL) {
    /* dumps: id node1 node2 parent_curve */
    ep->cell = ++eid;
    if (ep->parent_curve == NULL) {
      fprintf(fp, "%12lu %12lu %12lu %12d\n",
              ep->cell, ep->vertex_1->id, ep->vertex_2->id, 0);
    } else {
      fprintf(fp, "%12lu %12lu %12lu %12d\n",
              ep->cell, ep->vertex_1->id, ep->vertex_2->id,
              get_curve_id(ep->parent_curve));
    }
    ep = (EDGE_REC *)GET_LIST_NEXT(edge_list, ep);
  }
}


void
destroy_edge(EDGE_REC *e_p)
{
  remove_vert_reference(e_p->vertex_1, e_p);
  remove_vert_reference(e_p->vertex_2, e_p);
  DELETE_LIST_NODE(edge_list, e_p);
#ifdef X_XFACE
  if (options.X_xface) {
    EMUnlinkGraphics(ESIModel (), e_p->g_p);
  }
#endif
}
