/*  
    =========================================================================

    targe2: a program for "T"riangulation of "AR"bitrary "GE"ometries in "2"D

    =========================================================================

                      Copyright: 1993, 1994 Petr Krysl 

       Czech Technical University in Prague, Faculty of Civil Engineering,
           Dept. Structural Mechanics, 166 29 Prague, Czech Republic,
                          e-mail: pk@aix.fsv.cvut.cz

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    =========================================================================
*/
 
#include "targe2cf.h"
#include "ttypes.h"

#include <stdio.h>
#include <stdlib.h>
#ifdef UNIX
#include <unistd.h>
#endif
#include <string.h>
#include <errno.h>
#include <ctype.h>

#include "targe2.h"
#include "tops.h"
#include "log.h"
#include "input.h"
#include "cmds.h"
#include "utils.h"

 /* First the commands to be handled  */

static top_command   commands[] = {
  { NULL,     0,       error_in_input },
  { "MC-T",   4,       def_by_triangle_control },
  { "CURV",   4,       def_curve },
  { "MC-P",   4,       def_plain_ctl_point },
  { "FIXED",  5,       def_fixed_point },
  { "SUBR",   4,       def_subregion },
  { "MC-C",   4,       def_constant_mesh_size },
  { "M-CTL",  5,       def_mctl_point },
  { "MC-F",   4,       def_by_triangle_file },
  { "PRINT",  5,       print_cmd },
  { "STOP",   4,       stop_run },
  { "REGEXT", 6,       set_region_extent },
};

static int commands_total;

 /* FILE pointer for input */
static FILE *in_FILE_p = NULL;


static line_contents command_line;

static void open_in_file(void);



void 
input_data(void) 
{
  init_command_line();
  open_in_file();
  while(read_next_line() != EOF) {
    /*print_line_contents();*/
    commands[current_command()].handle_it();
  }
}




int 
current_command(void)
{
  int     i;
  char *tok;

  tok = get_token_as_string(1);

  for(i = 1; i < commands_total; i++) {
    if (compare_token_with_keyword(tok, commands[i].keyword) == YES)
      return i;
  }
  return 0;
}



void 
init_command_line(void)
{
  int i;
  
  for (i = 0; i < MAX_TOKENS; i++) {
    command_line.line_number                = 0;
    command_line.tokens[i].token_type       = NO_TOKEN;
    command_line.tokens[i].val              = NULL;
  }
  commands_total = sizeof(commands) / sizeof(top_command);
}



int 
read_next_line(void) /* returns number of tokens read from the line */
		     /* or EOF */
{
  char buffer[MAX_LINE_LENGTH];
  int anything_to_return = NO;

  while(anything_to_return == NO) {
    command_line.line_number++;
    if (fgets(buffer, MAX_LINE_LENGTH-1, in_FILE_p) == NULL) {
      if (feof(in_FILE_p))
	return EOF;
    }
    anything_to_return = parse_line(buffer);
  }
  return YES;
}



static void 
open_in_file(void) 
{
  if ((in_FILE_p = fopen(options.in_file_name, "r")) == NULL) {
    fprintf(stderr, "\nFailed to open input file \"%s\".",
	    options.in_file_name);
    exit(1);
  }
}




static BOOLEAN to_be_continued = NO;

#define LAST_TOKEN_IS_STRING \
  (command_line.tokens[command_line.tokens_total-1].token_type == STRING)

#define LINE_TO_BE_CONTINUED                                                \
 (strncmp(command_line.tokens[command_line.tokens_total-1].val, \
	  CONTINUATION_SEPARATOR, 2) == 0) 

int 
parse_line(char *buffer)
{
  char           *tok;
  BOOLEAN have_input = NO;
  char           *ptr;
  double      num_val;
  int          tok_ix;
  
  if (strlen(buffer) <= 1)
    return NO;
  
  tok         = strtok(buffer, SEPARATORS);
  if (*tok == COMMENT_CHAR)
    return NO;  /* first token is comment char -- return  */

  if (!to_be_continued)
    tok_ix      = NO_TOKENS;
  else
    tok_ix      = command_line.tokens_total-1;
  
  while (tok != NULL) {
    tok_ix++;
    if (tok_ix == MAX_TOKENS) {
      manip_error("Too many tokens in input line");
      break;
    }
    if ((command_line.tokens[tok_ix].val) != NULL) {
      free(command_line.tokens[tok_ix].val);
      command_line.tokens[tok_ix].val = NULL;
    }
    command_line.tokens[tok_ix].token_type      = NO_TOKEN;
    
    if (*tok == COMMENT_CHAR)
      goto postprocess;  /* this token is comment char -- return  */

    DUPLICATE_STRING(command_line.tokens[tok_ix].val, tok);
    command_line.tokens[tok_ix].token_type      = STRING;
    errno   = 0;
    num_val = strtod(tok, &ptr);
    if (ptr != tok && errno == 0 && fabs(num_val) <= FLT_MAX) {
      command_line.tokens[tok_ix].token_type      = NUMBER;
    }
    command_line.tokens_total = tok_ix + 1;
    have_input = YES;
    tok        = strtok(NULL, SEPARATORS);
  }
 postprocess:
  if (LAST_TOKEN_IS_STRING) {
    if (LINE_TO_BE_CONTINUED) {
      to_be_continued = YES;
      command_line.tokens_total -= 1; tok_ix = command_line.tokens_total;
      free(command_line.tokens[tok_ix].val);
      command_line.tokens[tok_ix].val = NULL;
      have_input = NO;
    }
  } else {
    to_be_continued = NO; 
  }
  return have_input;
}




int 
get_tokens_total(void)
{
  return command_line.tokens_total;
}



token_type 
get_token_type(int token_number)
{
  if (token_number <= command_line.tokens_total)
    return command_line.tokens[token_number-1].token_type;
  else
    return NO_TOKEN;
}



char *
get_token_as_string(int token_number)
{
  if (token_number <= command_line.tokens_total)
    return command_line.tokens[token_number-1].val;
  else
    return NULL;
}



double 
get_token_as_double(int token_number)
{
  char           *ptr;
  
  if (token_number <= command_line.tokens_total)    {
    if (command_line.tokens[token_number-1].token_type == NUMBER)
      return strtod(command_line.tokens[token_number-1].val, &ptr);
  }
  return 0.0;
}     



int 
bad_token(int ith, char *should_be_type)
{
  fprintf(stderr, "\nBad token type (should be a %s) in line #%d:\n "
	  "   %dth token = \"%s\".\n", should_be_type,
	  command_line.line_number,
	  ith, get_token_as_string(ith));
  exit(1);
  return 1;	/* to make compiler :-) */
}

	
#ifdef NEED_PRINT
void 
print_line_contents(void)
{
  int tot, i;
  token_type type;
  
  printf("\nLine # %d --> number of tokens: %d",
	 command_line.line_number, tot=get_tokens_total());
  for (i=1; i <= tot; i++)    {
    type = get_token_type(i);
    switch((int)type)	{
    case (int)NUMBER:
      printf("\n  %d type = NUMBER, val = %f",i,get_token_as_double(i));
      break;
    case (int)STRING:
      printf("\n  %d type = STRING, val = %s",i,get_token_as_string(i));
      break;
    default:
      printf("\n  %d type=UNKNOWN",i);
    }
  }
}
#endif


int 
is_keyword_in_line(int start_at, char *keyword)
{
  int i, tot;
  char  *tok;

  for(i = (start_at > 0)?start_at:1,
      tot = get_tokens_total(); i <= tot; i++) {
    tok = get_token_as_string(i);
    if (compare_token_with_keyword(tok, keyword) == YES)
      return i;
  }
  return 0;
}



void 
error_in_input(void)
{
  fprintf(stderr, "\nUnknown top keyword \"%s\" in line#%d. Ignoring.\n",
	  get_token_as_string(1), command_line.line_number);
}




BOOLEAN 
compare_token_with_keyword(char *tok, char *keyword)
{
  int j, len;

  if (tok == NULL)
    return NO;
  
  for(j = 0, len = min(strlen(tok), strlen(keyword)); j < len; j++) {
    if (toupper(*(tok+j)) != *(keyword+j))	{
      return NO;
    }
  }
  return YES;
}



void 
bad_syntax(char *to_print)
{
  fprintf(stderr, "\nBad syntax in line#%d. %s\n",
	  command_line.line_number, to_print);
  exit(1);
}
