/*  
    =========================================================================

    targe2: a program for "T"riangulation of "AR"bitrary "GE"ometries in "2"D

    =========================================================================

                      Copyright: 1993, 1994 Petr Krysl 

       Czech Technical University in Prague, Faculty of Civil Engineering,
           Dept. Structural Mechanics, 166 29 Prague, Czech Republic,
                          e-mail: pk@aix.fsv.cvut.cz

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    =========================================================================
*/

#include "targe2cf.h"
#include "ttypes.h"
#include "utils.h"
#include "tops.h" 
#include "mctlp.h"
#include "qtree.h"

#ifdef X_XFACE

#include <stdio.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xresource.h>
#include <X11/Intrinsic.h>

#define HAVE_BOOLEAN

#include "xiface.h"

#endif


 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */
 /* PRIVATE */
 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

extern OPTS_REC options; /* global options structure */

static MCTLP_LIST mctlp_list;

static float  const_cdim  = 0.0;   /* constant mesh-size dimension */
static float  min_cdim    = 1.e30; /* extremes of the characteristic */
static float  max_cdim    = 0.0;   /* dimension */
static float  exponent_mu = 2.0;
static float  inv_phi_bar = 1.1;

#ifdef MCTL_BY_FUNC
float mctl_func(float at_x, float at_y);


#define MIN_X (-100.)
#define MAX_X (+100.)
#define MIN_Y (-100.)
#define MAX_Y (+100.)
#endif
DBG(
void mctlp_info(OBJECT_P op, struct _GRAPHICS_REC *gp, MESSAGE_REC *mp);
)

#define TRI_MIN_MAX(xmin, xmax, x1, x2, x3) { \
					    xmin = min(x1, x2); \
					    xmin = min(xmin, x3); \
					    xmax = max(x1, x2); \
					    xmax = max(xmax, x3); \
					  }
#define FIRST_LEVEL_RATIO  0.25
#define SECOND_LEVEL_RATIO 0.1

static void 
mctlp_at_center(
                float d, 
                float x1, float y1,
                float x2, float y2,
                float x3, float y3,
                float v1, float v2, float v3,
                float mf1, float mf2, float mf3,
                float ang1, float ang2, float ang3);

static int mesh_control_available = NO;

#define TRIGGER(v) ((v) < d)

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */
 /* PUBLIC */
 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

void make_mctlp_list()
{
  mctlp_list  = make_list();
  assert(mctlp_list != NULL);
  exponent_mu = options.mctl_func_exponent; /* one-time initializations */
  inv_phi_bar = 1./SHAPE_FUNC_VAL_IN_INFL_RADIUS;
}

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

void add_mctl_point(int id, float x, float y, float cdim, float infl_radius)
{
  MCTLP_REC *mctlp_p;

#ifdef Q_TREE
  if (options.qtree) {
    qt_extend_qtree(x, y, cdim, 1, 0);
  } else {
#endif
    make_and_add_to_tail_val(mctlp_list, mctlp_p, MCTLP_REC, "add_mctl_point",
			     id);
    mctlp_p->coords.x    = x;
    mctlp_p->coords.y    = y;
    mctlp_p->near_cdim   = cdim;
    mctlp_p->infl_radius = infl_radius;
#ifdef Q_TREE
  }
#endif
  set_cdim_extremes(cdim);
  mesh_control_available = YES;
}

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */
#ifdef X_XFACE
#   define NEED_INPUT_CONTROL
#endif

void 
add_mctl_triangle(float x1, float y1,
                  float x2, float y2,
                  float x3, float y3,
                  float v1, float v2, float v3,
                  float mf1, float mf2, float mf3,
                  float ang1, float ang2, float ang3)
{
  float xc, yc, vc, xmin, xmax, ymin, ymax, dx, dy, vmin, vmax, d;
#ifdef NEED_INPUT_CONTROL
  WCRec pnts[3];
  static EPixel color = 0x0, blue = 0x0, red = 0x0;
  GraphicObj *g_p;
  BOOLEAN status;
#endif
  
  if (mf1 < 1 || mf2 < 1 || mf3 < 1)
    manip_error("Invalid anisotropic multiplier");

  /* Angles should be >= zero */
  while (ang1 < 0)
    ang1 += M_PI;
  while (ang2 < 0)
    ang2 += M_PI;
  while (ang3 < 0)
    ang3 += M_PI;
  
  /* Adjust the angles; we can do with angles in the interval <0;pi> */
  while (ang1 > M_PI)
    ang1 -= M_PI;
  while (ang2 > M_PI)
    ang2 -= M_PI;
  while (ang3 > M_PI)
    ang3 -= M_PI;
  
#ifdef Q_TREE
    if (options.qtree) {
#ifdef NEED_INPUT_CONTROL
      if (options.X_xface && options.show_on_Z) {
        if (color == 0x0) color = ColorGetPixelFromString("black", &status);
        if (   options.show_on_Z == 1
            || options.show_on_Z == 2
            || options.show_on_Z == 3) {
          pnts[0].x = x1; pnts[0].y = y1;
          pnts[1].x = x2; pnts[1].y = y2;
          pnts[2].x = x3; pnts[2].y = y3;
          if        (options.show_on_Z == 1) {
            pnts[0].z = v1; pnts[1].z = v2; pnts[2].z = v3;
          } else if (options.show_on_Z == 2) {
            pnts[0].z = mf1 * Z_SHOW_SCALE;
            pnts[1].z = mf2 * Z_SHOW_SCALE;
            pnts[2].z = mf3 * Z_SHOW_SCALE;
          } else if (options.show_on_Z == 3) {
            pnts[0].z = ang1 * Z_SHOW_SCALE;
            pnts[1].z = ang2 * Z_SHOW_SCALE;
            pnts[2].z = ang3 * Z_SHOW_SCALE;
          }
          g_p = CreateTriangle3D(pnts);
          EASValsSetColor(color);
          EGWithMaskChangeAttributes(WIDTH_MASK | COLOR_MASK, g_p);
          EMAddGraphicsToModel(ESIModel (), g_p);
        } else if (options.show_on_Z == 4) {
          if (blue == 0x0) blue = ColorGetPixelFromString("blue", &status);
          if (red == 0x0) red = ColorGetPixelFromString("red", &status);
#undef SHOW_DIR
#ifdef SHOW_DIR
          pnts[0].x = x1;
          pnts[0].y = y1;
#else
          pnts[0].x = x1 - 0.5*v1*cos(ang1);
          pnts[0].y = y1 - 0.5*v1*sin(ang1);
#endif
          pnts[0].z = 0;
          pnts[1].x = x1 + 0.5*v1*cos(ang1);
          pnts[1].y = y1 + 0.5*v1*sin(ang1);
          pnts[1].z = 0;
          g_p = CreateLine3D(pnts);
          EASValsSetColor(blue);
          EGWithMaskChangeAttributes(COLOR_MASK, g_p);
          EMAddGraphicsToModel(ESIModel (), g_p);
#ifdef SHOW_DIR
          pnts[0].x = x2;
          pnts[0].y = y2;
#else
          pnts[0].x = x2 - 0.5*v2*cos(ang2);
          pnts[0].y = y2 - 0.5*v2*sin(ang2);
#endif
          pnts[0].z = 0;
          pnts[1].x = x2 + 0.5*v2*cos(ang2);
          pnts[1].y = y2 + 0.5*v2*sin(ang2);
          pnts[1].z = 0;
          g_p = CreateLine3D(pnts);
          EASValsSetColor(blue);
          EGWithMaskChangeAttributes(COLOR_MASK, g_p);
          EMAddGraphicsToModel(ESIModel (), g_p);
          pnts[0].x = x3 - 0.5*v2*cos(ang3);
          pnts[0].y = y3 - 0.5*v2*sin(ang3);
          pnts[0].z = 0;
          pnts[1].x = x3 + 0.5*v2*cos(ang3);
          pnts[1].y = y3 + 0.5*v2*sin(ang3);
          pnts[1].z = 0;
          g_p = CreateLine3D(pnts);
          EASValsSetColor(blue);
          EGWithMaskChangeAttributes(COLOR_MASK, g_p);
          EMAddGraphicsToModel(ESIModel (), g_p);
          
#ifdef SHOW_DIR
          pnts[0].x = x1;
          pnts[0].y = y1;
#else
          pnts[0].x = x1 + 0.5*mf1*v1*sin(ang1);
          pnts[0].y = y1 - 0.5*mf1*v1*cos(ang1);
#endif
          pnts[0].z = 0;
          pnts[1].x = x1 - 0.5*mf1*v1*sin(ang1);
          pnts[1].y = y1 + 0.5*mf1*v1*cos(ang1);
          pnts[1].z = 0;
          g_p = CreateLine3D(pnts);
          EASValsSetColor(red);
          EGWithMaskChangeAttributes(COLOR_MASK, g_p);
          EMAddGraphicsToModel(ESIModel (), g_p);
          pnts[0].x = x2 + 0.5*mf2*v2*sin(ang2);
          pnts[0].y = y2 - 0.5*mf2*v2*cos(ang2);
          pnts[0].z = 0;
          pnts[1].x = x2 - 0.5*mf2*v2*sin(ang2);
          pnts[1].y = y2 + 0.5*mf2*v2*cos(ang2);
          pnts[1].z = 0;
          g_p = CreateLine3D(pnts);
          EASValsSetColor(red);
          EGWithMaskChangeAttributes(COLOR_MASK, g_p);
          EMAddGraphicsToModel(ESIModel (), g_p);
#ifdef SHOW_DIR
          pnts[0].x = x3;
          pnts[0].y = y3;
#else
          pnts[0].x = x3 + 0.5*mf3*v3*sin(ang3);
          pnts[0].y = y3 - 0.5*mf3*v3*cos(ang3);
#endif
          pnts[0].z = 0;
          pnts[1].x = x3 - 0.5*mf3*v3*sin(ang3);
          pnts[1].y = y3 + 0.5*mf3*v3*cos(ang3);
          pnts[1].z = 0;
          g_p = CreateLine3D(pnts);
          EASValsSetColor(red);
          EGWithMaskChangeAttributes(COLOR_MASK, g_p);
          EMAddGraphicsToModel(ESIModel (), g_p);
        }
      }
#endif
    qt_extend_qtree(x1, y1, v1, mf1, ang1);
    qt_extend_qtree(x2, y2, v2, mf2, ang2);
    qt_extend_qtree(x3, y3, v3, mf3, ang3);
    TRI_MIN_MAX(xmin, xmax, x1, x2, x3);
    TRI_MIN_MAX(ymin, ymax, y1, y2, y3);
    dx = xmax-xmin; dy = ymax-ymin; d = (dx+dy)/2;
    mctlp_at_center(d, x1, y1, x2, y2, x3, y3,
                    v1, v2, v3, mf1, mf2, mf3, ang1, ang2, ang3);
  } else {
#endif
    xc   = ONE_THIRD * (x1+x2+x3); 
    yc   = ONE_THIRD * (y1+y2+y3);
    vc   = ONE_THIRD * (v1+v2+v3);
    TRI_MIN_MAX(xmin, xmax, x1, x2, x3);
    TRI_MIN_MAX(ymin, ymax, y1, y2, y3);
    TRI_MIN_MAX(vmin, vmax, v1, v2, v3);
    dx = xmax-xmin; dy = ymax-ymin; d = 0.5*(dx+dy)*SQRT_2;
    add_mctl_point(0, xc, yc, vc, 0.5 * d);
#ifdef Q_TREE
  }
#endif
  set_cdim_extremes(v1);
  set_cdim_extremes(v2);
  set_cdim_extremes(v3);
  mesh_control_available = YES;
}

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

#ifdef X_XFACE
DBG(
void mctlp_info(OBJECT_P op, struct _GRAPHICS_REC *gp, MESSAGE_REC *mp)
{
  printf("MCTLP size %f\n", *((float *)op));
} )
#endif

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

#ifdef Q_TREE
static void 
mctlp_at_center(float d, 
                float x1, float y1,
                float x2, float y2,
                float x3, float y3,
                float v1, float v2, float v3,
                float mf1, float mf2, float mf3,
                float ang1, float ang2, float ang3)
{
  float xc, yc, a1, a2, a3, a;

#define AVG(VAL1, VAL2) (0.5*((VAL1)+(VAL2)))
#define AVGA(A1, A2)    AVG(A1, A2)
#define CORRECT(A) /* VOID */

#define EPS 0.05

  /* catch those cases, where the angles on a triangle differ by approx. PI */
  /* force the angles then to the same angle */
  if (fabs(fabs(ang1 - ang2) - M_PI) < EPS) {
    if (ang1 > ang2) { ang2 = ang1; }
    else             { ang1 = ang2; }
  }
  if (fabs(fabs(ang1 - ang3) - M_PI) < EPS) {
    if (ang1 > ang3) { ang3 = ang1; }
    else             { ang1 = ang3; }
  }
  if (fabs(fabs(ang2 - ang3) - M_PI) < EPS) {
    if (ang2 > ang3) { ang3 = ang2; }
    else             { ang2 = ang3; }
  }

  xc   = ONE_THIRD * (x1+x2+x3);
  yc   = ONE_THIRD * (y1+y2+y3);
  if (TRIGGER(v1)) {
    a1 = ang1;
    a2 = AVGA(ang1, ang2); CORRECT(a2);
    a3 = AVGA(ang1, ang3); CORRECT(a3);
    mctlp_at_center(0.5*d,
                    x1, y1, AVG(x1,x2), AVG(y1,y2), AVG(x1,x3), AVG(y1,y3),
                    v1, AVG(v1,v2), AVG(v1,v3),
                    mf1, AVG(mf1,mf2), AVG(mf1,mf3),
                    a1, a2, a3);
  }
  if (TRIGGER(v2)) {
    a1 = ang2;
    a2 = AVGA(ang1, ang2); CORRECT(a2);
    a3 = AVGA(ang2, ang3); CORRECT(a3);
    mctlp_at_center(0.5*d,
                    x2, y2, AVG(x1,x2), AVG(y1,y2), AVG(x2,x3), AVG(y2,y3),
                    v2, AVG(v1,v2), AVG(v2,v3),
                    mf2, AVG(mf1,mf2), AVG(mf2,mf3),
                    a1, a2, a3);
  }
  if (TRIGGER(v3)) {
    a1 = ang3;
    a2 = AVGA(ang1, ang3); CORRECT(a2);
    a3 = AVGA(ang2, ang3); CORRECT(a3);
    mctlp_at_center(0.5*d,
                    x3, y3, AVG(x1,x3), AVG(y1,y3), AVG(x2,x3), AVG(y2,y3),
                    v3, AVG(v1,v3), AVG(v2,v3),
                    mf3, AVG(mf1,mf3), AVG(mf2,mf3),
                    a1, a2, a3);
  }
  if (TRIGGER(AVG(v1,v2)) || TRIGGER(AVG(v1,v3)) || TRIGGER(AVG(v2,v3))) {
    a1 = AVGA(ang1, ang2); CORRECT(a1);
    a2 = AVGA(ang2, ang3); CORRECT(a2);
    a3 = AVGA(ang1, ang3); CORRECT(a3);
    a  = AVGA(a1, a2);     CORRECT(a);
    a  = AVGA(a, a3);      CORRECT(a);
    qt_extend_qtree(xc, yc, ONE_THIRD*(v1+v2+v3), ONE_THIRD*(mf1+mf2+mf3), a);
    mctlp_at_center(0.5*d,
		    AVG(x1,x2), AVG(y1,y2),
		    AVG(x2,x3), AVG(y2,y3),
		    AVG(x1,x3), AVG(y1,y3),
		    AVG(v1,v2), AVG(v2,v3), AVG(v1,v3),
                    AVG(mf1,mf2), AVG(mf2,mf3), AVG(mf1,mf3),
                    a1, a2, a3);
  }
}
#endif



void 
add_const_cdim(float cdim)
{
  const_cdim = cdim;
  set_cdim_extremes(cdim);
  mesh_control_available = YES;
}





void 
set_cdim_extremes(float cdim)
{
  min_cdim = min(cdim, min_cdim);
  max_cdim = max(cdim, max_cdim);
}



void 
get_cdim_extremes(float *min_cd, float *max_cd)
{
  *min_cd = min_cdim;
  *max_cd = max_cdim;
}



static double gh_beta = 2;
static double gh_beta_squared = 4;
static double gh_exp_neg_beta_squared = 0.01831563888873418029; /* = exp(-4) */
static double gh_A
            = 1.01865736036377404793; /* = 1/(1 - gh_exp_neg_beta_squared) */
static double gh_minus_two_A_beta_squared
            = -8.14925888291019238344;

/*
   Set parameters of the Gaussian hill weight.
   Set the beta parameter.  See the writeup for the Gaussian hill
   weight function for its meaning.
 */
void
gh_init (double beta)
{
  gh_beta                     = beta;
  gh_beta_squared             = gh_beta * gh_beta;
  gh_exp_neg_beta_squared     = exp(-gh_beta_squared);
  gh_A                        = 1 / (1 - gh_exp_neg_beta_squared);
  gh_minus_two_A_beta_squared = -2 * gh_A * gh_beta_squared;
}


/*
   The truncated Gaussian hill weight function.
   The parameter r is the distance normalized by the support size (support
   radius).
   Definition:
 " w(d_I) = (exp(-(d_I/c)^2) - exp(-(d_mI/c)^2)) / (1 - exp(-(d_mI/c)^2))
   with 
 " beta = d_mI/c, 
 " d_I/c = beta d_I/d_mI = beta r
   or alternatively, with the above definitions
 " w(d_I) = (exp(-(beta * r)^2) - exp(-beta^2)) / (1 - exp(-beta^2))
 */
float                              /* OUT weight function */
gh_W (
             float rx /* x-component of r */,
             float ry /* y-component of r */
)
{
  double r2 = ((rx) * (rx) + (ry) * (ry));
  
  if (r2 >= 1)
    return 0;
  else {
    return gh_A * (exp(-gh_beta_squared * r2) - gh_exp_neg_beta_squared);  
  }
}


static float 
W (float at_x, float at_y, MCTLP_REC *mp)
{
  static int init = 0;
  float R = 2*mp->infl_radius;
  float rx = (at_x - mp->coords.x) / (R);
  float ry = (at_y - mp->coords.y) / (R);

  if (!init) { gh_init (2.0); init = 1; }

  return 1.0e2 * gh_W (rx, ry);
}


static float 
strict (float at_x, float at_y)
{
  MCTLP_REC *mp;
  float mc_pu, sum_pu, inv_sum_pu, ms = 0;
#define LOOP_S()                                                      \
  mp      = NULL;                                                     \
  while((mp = (MCTLP_REC *)get_list_next(mctlp_list, mp)) != NULL) 

  sum_pu = 0;
  LOOP_S () {
    mp->pu = W (at_x, at_y, mp);
    sum_pu += mp->pu;
  }
  if (const_cdim > 0) sum_pu += 1;
  inv_sum_pu = 1 / sum_pu;
  LOOP_S () {
    mp->pu *= inv_sum_pu;
    sum_pu += mp->pu;
  }
  if (const_cdim > 0) mc_pu = 1 * inv_sum_pu;
  ms = 0;
  LOOP_S () {
    ms += mp->pu * mp->near_cdim;
  }
  if (const_cdim > 0) ms += mc_pu * const_cdim;

  return ms;
}


float 
get_mesh_cdim(float at_x, float at_y)
{
  float inflcoeffsum, dim, cradius, tmp, cinflc, weight;
  MCTLP_REC *mctlp_p;

  if (!mesh_control_available)
    manip_error("No mesh control defined yet!!!");
  
#ifdef Q_TREE
  if (options.qtree) {
    return qt_leaf_cdim_at(at_x, at_y);
  }
#endif
#ifdef MCTL_BY_FUNC
   if (options.mctl_by_func) {
    return mctl_func(at_x, at_y);
  } 
#endif    
#ifdef STRICT_MESH_SIZE_CONTROL
  if (options.strict_mesh_size_control) {
    return strict (at_x, at_y);
  }
#endif
  dim          = 0.0;
  inflcoeffsum = 0.0;
  
  mctlp_p      = (MCTLP_REC *)get_list_next(mctlp_list, NULL);
  
  while(mctlp_p != NULL)    {
    tmp      = (at_x - mctlp_p->coords.x);
    cradius  = tmp * tmp;
    tmp      = (at_y - mctlp_p->coords.y);
    cradius  = sqrt(cradius + tmp * tmp);
    
    if (mctlp_p->near_cdim > 0.0) {
      cinflc       = mctl_near_fval(cradius, mctlp_p->infl_radius);
      weight       = max_cdim / mctlp_p->near_cdim;
      inflcoeffsum += cinflc * weight;
      dim          += cinflc * max_cdim;
    }
    
    mctlp_p          = (MCTLP_REC *)get_list_next(mctlp_list, mctlp_p);
  }
  if (const_cdim > 0.0) 
    return (dim + const_cdim) / (inflcoeffsum + 1.0);
  else
    return (dim / inflcoeffsum);
}




float 
get_mesh_2nd_dir_magn_factor(float at_x, float at_y)
{
  if (!mesh_control_available)
    manip_error("No mesh control defined yet!!!");
  
#ifdef Q_TREE
  if (options.qtree) {
    return qt_2nd_dir_magn_factor_at(at_x, at_y);
  }
#endif
    return 1; /* Default behaviour -- isotropic mesh */
}



float 
get_mesh_1st_dir_angle(float at_x, float at_y)
{
  if (!mesh_control_available)
    manip_error("No mesh control defined yet!!!");
  
#ifdef Q_TREE
  if (options.qtree) {
    return qt_1st_dir_angle_at(at_x, at_y);
  }
#endif
    return 0; /* Default behaviour -- isotropic mesh */
}




float 
mctl_near_fval(float radius, float infl_radius)
{
  float rho;

  rho = radius / infl_radius;
  return (float)(1. / (1. + pow(rho, exponent_mu) * (inv_phi_bar - 1.)));
}



void 
get_maxmin_mesh_cdim(float xmin, float ymin, float xmax, float ymax,
                     float *minmeshsize, float *maxmeshsize)
{
  MCTLP_REC *min_p_p = NULL, *max_p_p = NULL, *curr_p_p;
  float dim;

#ifdef Q_TREE
  if (options.qtree) {
    *minmeshsize = min_cdim;
    *maxmeshsize = max_cdim;
    goto exit;
  } 
#endif 
#ifdef MCTL_BY_FUNC
  if (options.mctl_by_func) {
    float tmp, x, y, step;

    step = min(((MAX_X)-(MIN_X)), ((MAX_Y)-(MIN_Y))); step /= 100.;
    *minmeshsize = 1.e38; *maxmeshsize = 0;
    for (x = MIN_X; x < MAX_X; x += step)
      for (y = MIN_Y; y < MAX_Y; y += step) {
	tmp = mctl_func(x, y);
	*minmeshsize = min(tmp, *minmeshsize);
	*maxmeshsize = max(tmp, *maxmeshsize);
      }
    goto exit;
  } 
#endif
  dim = get_mesh_cdim(xmin, ymin);
  *minmeshsize = dim;
  *maxmeshsize = dim;
  dim = get_mesh_cdim(xmin, ymax);
  *minmeshsize = min(*minmeshsize, dim);
  *maxmeshsize = max(*maxmeshsize, dim);
  dim = get_mesh_cdim(xmax, ymin);
  *minmeshsize = min(*minmeshsize, dim);
  *maxmeshsize = max(*maxmeshsize, dim);
  dim = get_mesh_cdim(xmax, ymax);
  *minmeshsize = min(*minmeshsize, dim);
  *maxmeshsize = max(*maxmeshsize, dim);

  if (list_empty(mctlp_list))
    if (const_cdim <= 0.0)
      manip_error("No mesh control information!?");
  
#define IN_LIMITS(v, minv, maxv) ((v) >= (minv) && (v) <= (maxv))
  
  curr_p_p = (MCTLP_REC *)get_list_next(mctlp_list, NULL);
  while (curr_p_p != NULL) {
    if (IN_LIMITS(curr_p_p->coords.x, xmin, xmax) &&
	IN_LIMITS(curr_p_p->coords.y, ymin, ymax)) {
      if (curr_p_p->infl_radius > 0.) {
	if (curr_p_p->near_cdim < *minmeshsize) {
	  *minmeshsize = curr_p_p->near_cdim;
	  min_p_p      = curr_p_p;
	}
	if (curr_p_p->near_cdim > *maxmeshsize) {
	  *maxmeshsize = curr_p_p->near_cdim;
	  max_p_p      = curr_p_p;
	}
      }
    }
    curr_p_p = (MCTLP_REC *)get_list_next(mctlp_list, curr_p_p);
  }
  if (min_p_p != NULL) {
    *minmeshsize = get_mesh_cdim(min_p_p->coords.x, min_p_p->coords.y);
  }
  if (max_p_p != NULL) {
    *maxmeshsize = get_mesh_cdim(max_p_p->coords.x, max_p_p->coords.y);
  }
  if (const_cdim > 0.0) {
    *minmeshsize = min(*minmeshsize, const_cdim);
    *maxmeshsize = max(*maxmeshsize, const_cdim);
  }
 exit:
  ;
}


#ifdef MCTL_BY_FUNC
#error "How did I get here"
#define MINMS      0.0005
#define MINMINMS   0.00025
#define MAXMS      10.0

#define EVAL_NEAR_FVAL(a_NUM) { \
      c += mctl_near_fval(line_length(at_x, at_y, x##a_NUM, y##a_NUM), \
                          R##a_NUM); \
      w = maxs / s##a_NUM; sum += c * w; s += c * maxs; }


float mctl_func(float at_x, float at_y)
{
  int funcnum = options.mctl_by_func;

  switch (funcnum) {
  case 1:
    {
      float tmp, ksi, eta;
      
      ksi = at_x;
      eta = at_y;
      tmp = 0.5*fabs((eta+ksi)*sin(0.02*eta))+1;
      return tmp;
    }
    break;
  case 2:
    {
      float tmp, ksi, eta;
      
      ksi = (at_x-0.5*(MAX_X+MIN_X))/(MAX_X-MIN_X);
      eta = (at_y-0.5*(MAX_Y+MIN_Y))/(MAX_Y-MIN_Y);
      tmp = sqrt(1000*eta*eta*sqrt(10*fabs(sin(sqrt(ksi*ksi)))));
      tmp = fabs(tmp) + 5;
      return tmp;
    }
    break;
  case 3:
    {
      float tmp, ksi, eta;
      
      ksi = (at_x-0.5*(MAX_X+MIN_X))/(MAX_X-MIN_X);
      eta = (at_y-0.5*(MAX_Y+MIN_Y))/(MAX_Y-MIN_Y);
      tmp = fabs(sin(6*eta)*sin(6*ksi));
      tmp = 30*sqrt(tmp*tmp*tmp) + 0.5;
      return tmp;
    }
    break;
  case 4:
    {
      float tmp, ksi, eta, r;
      
      ksi = (at_x-0.5*(MAX_X+MIN_X))/(MAX_X-MIN_X);
      eta = (at_y-0.5*(MAX_Y+MIN_Y))/(MAX_Y-MIN_Y);
      r   = sqrt(ksi*ksi+eta*eta)+0.01;
      tmp = fabs(sin((eta*(-eta)*(-ksi))));
      tmp = 20*sqrt(fabs(ksi-eta)*tmp) + 0.5;
      return tmp/r/r;
    } 
    break;
  case 5: /* *** Fine circumference, gross center */
    {
      float ksi, eta, r;
      
      ksi = (at_x-0.5*(MAX_X+MIN_X))/(MAX_X-MIN_X);
      eta = (at_y-0.5*(MAX_Y+MIN_Y))/(MAX_Y-MIN_Y);
      r   = (ksi*ksi+eta*eta)+0.5;
      return 1/r/r/r/r/r;
    }
    break;
  case 6: /* *** Concentric circles of small size */
    {
      float ksi, eta, r;
      
      ksi = (at_x-0.5*(MAX_X+MIN_X))/(MAX_X-MIN_X);
      eta = (at_y-0.5*(MAX_Y+MIN_Y))/(MAX_Y-MIN_Y);
      r   = (ksi*ksi+eta*eta);
      return 8*(fabs(sin(24*r))+0.1);
    }
    break;
  case 7: /* *** */
    {
      float ksi, eta, r;
      
      ksi = (at_x-0.5*(MAX_X+MIN_X))/(MAX_X-MIN_X);
      eta = (at_y-0.5*(MAX_Y+MIN_Y))/(MAX_Y-MIN_Y);
      r   = (ksi*ksi+eta*eta);
      return ((at_x*(at_x+1))/100 + 5);
    }
    break;
  case 8:
    {
      float tmp, ksi, eta;
      
      ksi = (at_x-0.5*(MAX_X+MIN_X))/(MAX_X-MIN_X);
      eta = (at_y-0.5*(MAX_Y+MIN_Y))/(MAX_Y-MIN_Y);
      tmp = 30*fabs(sin(6*eta)*sin(6*ksi)) + 0.5;
      return tmp;
    }
    break;
  case 9:
    {
      float tmp, ksi, eta;
      
      ksi = (at_x-0.5*(MAX_X+MIN_X))/(MAX_X-MIN_X);
      eta = (at_y-0.5*(MAX_Y+MIN_Y))/(MAX_Y-MIN_Y);
      tmp = 50*fabs(sin(8*eta)*sin(12*ksi)) + 0.5;
      return tmp;
    }
    break;
  case 10:			/*  */
    {
      float x0 =-100,        y0 =-100,        s0 = MINMINMS, R0 = 0.002;
      float x1 = -47.297298, y1 = -15.000001, s1 = MINMS,    R1 = 0.002;
      float x2 =  47.027027, y2 = -22.297298, s2 = MINMS,    R2 = 0.002;
      float x3 =  19.189190, y3 = -59.054054, s3 = MINMS,    R3 = 0.002;
      float x4 = -39.189190, y4 = -56.621624, s4 = MINMS,    R4 = 0.002;
      float x5 =   8.918919, y5 =  76.351349, s5 = MINMS,    R5 = 0.002;
      float x6 =  16.011372, y6 = -17.911839, s6 = MINMS,    R6 = 0.002;
      float x7 =  16.722919, y7 = -16.857174, s7 = MINMS,    R7 = 0.002;
      float x8 =   1.413576, y8 =  63.765728, s8 = MINMS,    R8 = 0.002;
      float x9 =  40.914181, y9 =  -6.125616, s9 = MINMS,    R9 = 0.002;
      float x10= -69.189186, y10=  78.513512, s10= MINMS,    R10= 0.002;
      float maxs = MAXMS;
      float c = 0, sum = 0, w = 0, s = 0;

      EVAL_NEAR_FVAL(0);
      EVAL_NEAR_FVAL(1);
      EVAL_NEAR_FVAL(2);

      return (s + maxs) / (sum + 1);
    }
    break;
  case 11:			/*  */
    {
      float x0 =-100,        y0 =-100,        s0 = MINMINMS, R0 = 0.002;
      float x1 = -47.297298, y1 = -15.000001, s1 = MINMS,    R1 = 0.002;
      float x2 =  47.027027, y2 = -22.297298, s2 = MINMS,    R2 = 0.002;
      float x3 =  19.189190, y3 = -59.054054, s3 = MINMS,    R3 = 0.002;
      float x4 = -39.189190, y4 = -56.621624, s4 = MINMS,    R4 = 0.002;
      float x5 =   8.918919, y5 =  76.351349, s5 = MINMS,    R5 = 0.002;
      float x6 =  16.011372, y6 = -17.911839, s6 = MINMS,    R6 = 0.002;
      float x7 =  16.722919, y7 = -16.857174, s7 = MINMS,    R7 = 0.002;
      float x8 =   1.413576, y8 =  63.765728, s8 = MINMS,    R8 = 0.002;
      float x9 =  40.914181, y9 =  -6.125616, s9 = MINMS,    R9 = 0.002;
      float x10= -69.189186, y10=  78.513512, s10= MINMS,    R10= 0.002;
      float maxs = MAXMS;      
      float c = 0, sum = 0, w = 0, s = 0;

      EVAL_NEAR_FVAL(0);
      EVAL_NEAR_FVAL(1);
      EVAL_NEAR_FVAL(2);
      EVAL_NEAR_FVAL(3);
      EVAL_NEAR_FVAL(4);

      return (s + maxs) / (sum + 1);
    }
    break;
  case 12:			/* as 11 */
    {
      float x0 =-100,        y0 =-100,        s0 = MINMINMS, R0 = 0.002;
      float x1 = -47.297298, y1 = -15.000001, s1 = MINMS,    R1 = 0.002;
      float x2 =  47.027027, y2 = -22.297298, s2 = MINMS,    R2 = 0.002;
      float x3 =  19.189190, y3 = -59.054054, s3 = MINMS,    R3 = 0.002;
      float x4 = -39.189190, y4 = -56.621624, s4 = MINMS,    R4 = 0.002;
      float x5 =   8.918919, y5 =  76.351349, s5 = MINMS,    R5 = 0.002;
      float x6 =  16.011372, y6 = -17.911839, s6 = MINMS,    R6 = 0.002;
      float x7 =  16.722919, y7 = -16.857174, s7 = MINMS,    R7 = 0.002;
      float x8 =   1.413576, y8 =  63.765728, s8 = MINMS,    R8 = 0.002;
      float x9 =  40.914181, y9 =  -6.125616, s9 = MINMS,    R9 = 0.002;
      float x10= -69.189186, y10=  78.513512, s10= MINMS,    R10= 0.002;
      float maxs = MAXMS;      
      float c = 0, sum = 0, w = 0, s = 0;
      
      EVAL_NEAR_FVAL(0);
      EVAL_NEAR_FVAL(1);
      EVAL_NEAR_FVAL(2);
      EVAL_NEAR_FVAL(3);
      EVAL_NEAR_FVAL(4);
      EVAL_NEAR_FVAL(5);
      EVAL_NEAR_FVAL(6);
      EVAL_NEAR_FVAL(7);
      
      return (s + maxs) / (sum + 1);
    }
    break;
  case 13:			/* as 11 */
    {
      float x0 =-100,        y0 =-100,        s0 = MINMINMS, R0 = 0.002;
      float x1 = -47.297298, y1 = -15.000001, s1 = MINMS,    R1 = 0.002;
      float x2 =  47.027027, y2 = -22.297298, s2 = MINMS,    R2 = 0.002;
      float x3 =  19.189190, y3 = -59.054054, s3 = MINMS,    R3 = 0.002;
      float x4 = -39.189190, y4 = -56.621624, s4 = MINMS,    R4 = 0.002;
      float x5 =   8.918919, y5 =  76.351349, s5 = MINMS,    R5 = 0.002;
      float x6 =  16.011372, y6 = -17.911839, s6 = MINMS,    R6 = 0.002;
      float x7 =  16.722919, y7 = -16.857174, s7 = MINMS,    R7 = 0.002;
      float x8 =   1.413576, y8 =  63.765728, s8 = MINMS,    R8 = 0.002;
      float x9 =  40.914181, y9 =  -6.125616, s9 = MINMS,    R9 = 0.002;
      float x10= -69.189186, y10=  78.513512, s10= MINMS,    R10= 0.002;
      float maxs = MAXMS; 
      float c = 0, sum = 0, w = 0, s = 0;
      
      EVAL_NEAR_FVAL(0);
      EVAL_NEAR_FVAL(1);
      EVAL_NEAR_FVAL(2);
      EVAL_NEAR_FVAL(3);
      EVAL_NEAR_FVAL(4);
      EVAL_NEAR_FVAL(5);
      EVAL_NEAR_FVAL(6);
      EVAL_NEAR_FVAL(7);
      EVAL_NEAR_FVAL(8);
      EVAL_NEAR_FVAL(9);
      EVAL_NEAR_FVAL(10);
      
      return (s + maxs) / (sum + 1);
    }
    break;
  case 14:			/* as 11 */
    {
      float x0 =-100,        y0 =-100,        s0 = MINMINMS, R0 = 0.002;
      float x1 = -47.297298, y1 = -15.000001, s1 = MINMS,    R1 = 0.002;
      float x2 =  47.027027, y2 = -22.297298, s2 = MINMS,    R2 = 0.002;
      float x3 =  19.189190, y3 = -59.054054, s3 = MINMS,    R3 = 0.002;
      float x4 = -39.189190, y4 = -56.621624, s4 = MINMS,    R4 = 0.002;
      float x5 =   8.918919, y5 =  76.351349, s5 = MINMS,    R5 = 0.002;
      float x6 =  16.011372, y6 = -17.911839, s6 = MINMS,    R6 = 0.002;
      float x7 =  16.722919, y7 = -16.857174, s7 = MINMS,    R7 = 0.002;
      float x8 =   1.413576, y8 =  63.765728, s8 = MINMS,    R8 = 0.002;
      float x9 =  40.914181, y9 =  -6.125616, s9 = MINMS,    R9 = 0.002;
      float x10= -69.189186, y10=  78.513512, s10= MINMS,    R10= 0.002;
      float x11=   6.893034, y11=  27.161118, s11= MINMS,    R11= 0.002;
      float x12=   0.010973, y12=  20.013243, s12= MINMS,    R12= 0.002;
      float x13= -12.174173, y13=  29.700476, s13= MINMS,    R13= 0.002;
      float x14=  -3.267744, y14=  34.049507, s14= MINMS,    R14= 0.002;
      float maxs = MAXMS; 
      float c = 0, sum = 0, w = 0, s = 0;
      
      EVAL_NEAR_FVAL(0);
      EVAL_NEAR_FVAL(1);
      EVAL_NEAR_FVAL(2);
      EVAL_NEAR_FVAL(3);
      EVAL_NEAR_FVAL(4);
      EVAL_NEAR_FVAL(5);
      EVAL_NEAR_FVAL(6);
      EVAL_NEAR_FVAL(7);
      EVAL_NEAR_FVAL(8);
      EVAL_NEAR_FVAL(9);
      EVAL_NEAR_FVAL(10);
      EVAL_NEAR_FVAL(11);
      EVAL_NEAR_FVAL(12);
      EVAL_NEAR_FVAL(13);
      EVAL_NEAR_FVAL(14);
      
      return (s + maxs) / (sum + 1);
    }
    break;
  case 15:			/* as 11 */
    {
      float x0 =-100,        y0 =-100,        s0 = MINMINMS, R0 = 0.002;
      float x1 = -47.297298, y1 = -15.000001, s1 = MINMS,    R1 = 0.002;
      float x2 =  47.027027, y2 = -22.297298, s2 = MINMS,    R2 = 0.002;
      float x3 =  19.189190, y3 = -59.054054, s3 = MINMS,    R3 = 0.002;
      float x4 = -39.189190, y4 = -56.621624, s4 = MINMS,    R4 = 0.002;
      float x5 =   8.918919, y5 =  76.351349, s5 = MINMS,    R5 = 0.002;
      float x6 =  16.011372, y6 = -17.911839, s6 = MINMS,    R6 = 0.002;
      float x7 =  16.722919, y7 = -16.857174, s7 = MINMS,    R7 = 0.002;
      float x8 =   1.413576, y8 =  63.765728, s8 = MINMS,    R8 = 0.002;
      float x9 =  40.914181, y9 =  -6.125616, s9 = MINMS,    R9 = 0.002;
      float x10= -69.189186, y10=  78.513512, s10= MINMS,    R10= 0.002;
      float x11=   6.893034, y11=  27.161118, s11= MINMS,    R11= 0.002;
      float x12=   0.010973, y12=  20.013243, s12= MINMS,    R12= 0.002;
      float x13= -12.174173, y13=  29.700476, s13= MINMS,    R13= 0.002;
      float x14=  -3.267744, y14=  34.049507, s14= MINMS,    R14= 0.002;
      float x15=  88.648651, y15= -17.162163, s15= MINMS,    R15= 0.002;
      float x16= -21.621622, y16= -96.351349, s16= MINMS,    R16= 0.002;
      float x17= -37.297298, y17= -96.081085, s17= MINMS,    R17= 0.002;
      float x18= 100.000000, y18= 100.000000, s18= MINMS,    R18= 0.002;
      float x19=  -2.871984, y19=  0.807037,  s19= MINMS,    R19= 0.002;
      float maxs = MAXMS; 
      float c = 0, sum = 0, w = 0, s = 0;
      
      EVAL_NEAR_FVAL(0);
      EVAL_NEAR_FVAL(1);
      EVAL_NEAR_FVAL(2);
      EVAL_NEAR_FVAL(3);
      EVAL_NEAR_FVAL(4);
      EVAL_NEAR_FVAL(5);
      EVAL_NEAR_FVAL(6);
      EVAL_NEAR_FVAL(7);
      EVAL_NEAR_FVAL(8);
      EVAL_NEAR_FVAL(9);
      EVAL_NEAR_FVAL(10);
      EVAL_NEAR_FVAL(11);
      EVAL_NEAR_FVAL(12);
      EVAL_NEAR_FVAL(13);
      EVAL_NEAR_FVAL(14);
      EVAL_NEAR_FVAL(15);
      EVAL_NEAR_FVAL(16);
      EVAL_NEAR_FVAL(17);
      EVAL_NEAR_FVAL(18);
      EVAL_NEAR_FVAL(19);
      
      return (s + maxs) / (sum + 1);
    }
    break;
  case 24:
    return 2;
    break;
  case 25:
    return 1.5;
    break;
  case 26:
    return 1;
    break;
  case 27:
    return 0.85;
    break;
  case 28:
    return 0.6;
    break;
  case 29:
    return 0.5;
    break;
  case 30:
    return 0.45;
    break;
  case 31:
    return 0.4;
    break;
  case 32:
    return 0.35;
    break; 
  case 33:
    return 0.3;
    break; 
  default:
    {
      if (funcnum < 0)
	return 25;
      
      manip_error("No such control function number");
      return 0;
    }
    break;
  }
}
       

#endif
