/*  
    =========================================================================

    targe2: a program for "T"riangulation of "AR"bitrary "GE"ometries in "2"D

    =========================================================================

                      Copyright: 1993, 1994 Petr Krysl 

       Czech Technical University in Prague, Faculty of Civil Engineering,
           Dept. Structural Mechanics, 166 29 Prague, Czech Republic,
                          e-mail: pk@aix.fsv.cvut.cz

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    =========================================================================
*/

#include "targe2cf.h"
#include "ttypes.h"

#include <stdio.h>
#ifdef UNIX
#include <unistd.h>
#endif
#include <string.h>
#include <time.h>

#include "targe2.h"
#include "tops.h"
#include "log.h"

#include "subregs.h"
#include "vertex.h"
#include "edges.h"
#include "triangle.h"
#include "loops.h"
#include "mctlp.h"
#include "mesh.h"


VERTEX_LIST vertex_list;


void
remove_3_and_4_connectivity_vertices(void);

static void
reconstruct_3_patch(VERTEX_REC *v_p);
static void
reconstruct_4_patch(VERTEX_REC *v_p);

void
remove_3_and_4_connectivity_vertices(void)
{
  VERTEX_REC *v_p;

  v_p = (VERTEX_REC *)GET_LIST_NEXT(vertex_list, NULL);
  while (v_p != NULL) { /* loop over all vertices */
    if (v_p->parent_curve == NULL) {
      if (get_vertex_connectivity(v_p) == 3) {
        reconstruct_3_patch(v_p);
      } else if (get_vertex_connectivity(v_p) == 4) {
        reconstruct_4_patch(v_p);
      }
    }
    v_p = (VERTEX_REC *)GET_LIST_NEXT(vertex_list, v_p); 
  }                      /* loop over all vertices */
}

static void
reorder(VERTEX_REC *v1, VERTEX_REC **v2, VERTEX_REC **v3);
static void
reorder_4(float x, float y,
          VERTEX_REC **v1, VERTEX_REC **v2, VERTEX_REC **v3, VERTEX_REC **v4);
SUBREG_REC *
delete_triangles_and_edges_around_vertex(VERTEX_REC *v_p);

static void
reconstruct_3_patch(VERTEX_REC *v_p)
{
  VERTEX_REC *v1, *v2, *v3;
  EDGE_REC *e1, *e2, *e3, *ane;
  SUBREG_REC *subreg;
  EDGE_REC *er[MAX_CONNECTIVITY];
  int i, n;

  get_vert_references(v_p, &n, er);
  ane = er[0];
  v1 = ane->vertex_1;
  if (v1 == v_p)
    v1 = ane->vertex_2;
  ane = er[1];
  v2 = ane->vertex_1;
  if (v2 == v_p)
    v2 = ane->vertex_2;
  ane = er[2];
  v3 = ane->vertex_1;
  if (v3 == v_p)
    v3 = ane->vertex_2;
  reorder(v1, &v2, &v3);  /* order them counterclockwise */
  get_vert_references(v1, &n, er);
  for (i = 0; i < n; i++) {
    if (   (er[i]->vertex_1 == v1 && er[i]->vertex_2 == v2)
        || (er[i]->vertex_1 == v2 && er[i]->vertex_2 == v1)) {
      e1 = er[i];
    }
    if (   (er[i]->vertex_1 == v1 && er[i]->vertex_2 == v3)
        || (er[i]->vertex_1 == v3 && er[i]->vertex_2 == v1)) {
      e3 = er[i];
    }
  }
  get_vert_references(v2, &n, er);
  for (i = 0; i < n; i++) {
    if (   (er[i]->vertex_1 == v3 && er[i]->vertex_2 == v2)
        || (er[i]->vertex_1 == v2 && er[i]->vertex_2 == v3)) {
      e2 = er[i];
      break;
    }
  }
  
  subreg = delete_triangles_and_edges_around_vertex(v_p);
  add_triangle(e1, e2, e3, subreg);
  destroy_vertex(v_p);
}


static void
reorder(VERTEX_REC *v1, VERTEX_REC **v2, VERTEX_REC **v3)
{
  float x1, y1, x2, y2, x3, y3;
  float dx1, dy1, dx2, dy2;
  VERTEX_REC *v;

  get_vertex_coords(v1, &x1, &y1);
  get_vertex_coords(*v2, &x2, &y2);
  get_vertex_coords(*v3, &x3, &y3);
  dx1 = x2 - x1;
  dy1 = y2 - y1;
  dx2 = x3 - x1;
  dy2 = y3 - y1;
  if (dx1 * dy2 - dy1 * dx2 < 0) { /* z-component of cross product */
    v = *v2; *v2 = *v3; *v3 = v; /* swap them, if not counterclockwise */
  }
}

static void
add_triangles(VERTEX_REC *v1, VERTEX_REC *v2, VERTEX_REC *v3, VERTEX_REC *v4,
              EDGE_REC *e1, EDGE_REC *e2, EDGE_REC *e3, EDGE_REC *e4,
              SUBREG_REC *subreg);

static void
reconstruct_4_patch(VERTEX_REC *v_p)
{
  VERTEX_REC *v1, *v2, *v3, *v4;
  EDGE_REC *e1, *e2, *e3, *e4, *ane;
  SUBREG_REC *subreg;
  EDGE_REC *er[MAX_CONNECTIVITY];
  int i, n;
  float x, y;

  get_vertex_coords(v_p, &x, &y);
  get_vert_references(v_p, &n, er);
  ane = er[0];
  v1 = ane->vertex_1;
  if (v1 == v_p)
    v1 = ane->vertex_2;
  ane = er[1];
  v2 = ane->vertex_1;
  if (v2 == v_p)
    v2 = ane->vertex_2;
  ane = er[2];
  v3 = ane->vertex_1;
  if (v3 == v_p)
    v3 = ane->vertex_2;
  ane = er[3];
  v4 = ane->vertex_1;
  if (v4 == v_p)
    v4 = ane->vertex_2;
  reorder_4(x, y, &v1, &v2, &v3, &v4);  /* order them counterclockwise */
  get_vert_references(v1, &n, er);
  for (i = 0; i < n; i++) {
    if (   (er[i]->vertex_1 == v1 && er[i]->vertex_2 == v2)
        || (er[i]->vertex_1 == v2 && er[i]->vertex_2 == v1)) {
      e1 = er[i];
    }
    if (   (er[i]->vertex_1 == v1 && er[i]->vertex_2 == v4)
        || (er[i]->vertex_1 == v4 && er[i]->vertex_2 == v1)) {
      e4 = er[i];
    }
  }
  get_vert_references(v3, &n, er);
  for (i = 0; i < n; i++) {
    if (   (er[i]->vertex_1 == v3 && er[i]->vertex_2 == v2)
        || (er[i]->vertex_1 == v2 && er[i]->vertex_2 == v3)) {
      e2 = er[i];
    }
    if (   (er[i]->vertex_1 == v3 && er[i]->vertex_2 == v4)
        || (er[i]->vertex_1 == v4 && er[i]->vertex_2 == v3)) {
      e3 = er[i];
    }
  }
  
  subreg = delete_triangles_and_edges_around_vertex(v_p);
  add_triangles(v1, v2, v3, v4, e1, e2, e3, e4, subreg);
  destroy_vertex(v_p); 
}


static void
add_triangles(VERTEX_REC *v1, VERTEX_REC *v2, VERTEX_REC *v3, VERTEX_REC *v4,
              EDGE_REC *e1, EDGE_REC *e2, EDGE_REC *e3, EDGE_REC *e4,
              SUBREG_REC *subreg)
{
  float d1, d2;
  float x1, y1, x2, y2, x3, y3, x4, y4;
  EDGE_REC *e;

  get_vertex_coords(v1, &x1, &y1);
  get_vertex_coords(v2, &x2, &y2);
  get_vertex_coords(v3, &x3, &y3);
  get_vertex_coords(v4, &x4, &y4);
  
  d1 = line_length(x1, y1, x3, y3);
  d2 = line_length(x2, y2, x4, y4);

  if (d1 < d2) {
    e = add_edge(v1, v3, NULL);
    add_triangle(e1, e2, e, subreg);
    add_triangle(e3, e4, e, subreg);
  } else {
    e = add_edge(v2, v4, NULL);
    add_triangle(e4, e1, e, subreg);
    add_triangle(e2, e3, e, subreg);
  }
}



static void
reorder_4(float x, float y,
          VERTEX_REC **v1, VERTEX_REC **v2, VERTEX_REC **v3, VERTEX_REC **v4)
{
  float x1, y1, x2, y2, x3, y3, x4, y4;
  double t[5];
  int indx[5];
  VERTEX_REC *v;

  get_vertex_coords(*v1, &x1, &y1);
  get_vertex_coords(*v2, &x2, &y2);
  get_vertex_coords(*v3, &x3, &y3);
  get_vertex_coords(*v4, &x4, &y4);

#define SWAP_EM(v1, v2) { VERTEX_REC *tmp_v; tmp_v=(v1); (v1)=(v2); (v2)=tmp_v; }

#define SWAP_T(v1, v2) { double tmp_t; tmp_t=(v1); (v1)=(v2); (v2)=tmp_t; }

  t[1] = atan2(y1 - y, x1 - x);
  t[2] = atan2(y2 - y, x2 - x);
  t[3] = atan2(y3 - y, x3 - x);
  t[4] = atan2(y4 - y, x4 - x);
  if (t[2] < t[1]) { SWAP_EM(*v1, *v2); SWAP_T(t[1], t[2]);  }
  if (t[3] < t[2]) { SWAP_EM(*v3, *v2); SWAP_T(t[3], t[2]);  }
  if (t[4] < t[3]) { SWAP_EM(*v4, *v3); SWAP_T(t[4], t[3]);  }

  if (t[2] < t[1]) { SWAP_EM(*v1, *v2); SWAP_T(t[1], t[2]);  }
  if (t[3] < t[2]) { SWAP_EM(*v3, *v2); SWAP_T(t[3], t[2]);  }

  if (t[2] < t[1]) { SWAP_EM(*v1, *v2); SWAP_T(t[1], t[2]);  }
}
