/*  
    =========================================================================

    targe2: a program for "T"riangulation of "AR"bitrary "GE"ometries in "2"D

    =========================================================================

                      Copyright: 1993, 1994 Petr Krysl 

       Czech Technical University in Prague, Faculty of Civil Engineering,
           Dept. Structural Mechanics, 166 29 Prague, Czech Republic,
                          e-mail: pk@aix.fsv.cvut.cz

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    =========================================================================
*/

#include "targe2cf.h"
#include "ttypes.h"

#include <stdio.h>
#include <errno.h>
#ifdef SYSV
#include <unistd.h>
#endif
#include <string.h>
#include <time.h>

#include "log.h"
#include "tops.h"

#include "vertex.h"
#include "curve.h"
#include "loops.h"
#include "subregs.h"
#include "edges.h"
#include "mctlp.h"
#include "triangle.h"
#include "output.h"


#ifdef HAVE_STRERROR
#  if (!HAVE_STRERROR)
#    undef HAVE_STRERROR
#  endif
#endif
#ifndef HAVE_STRERROR
#     define strerror(E_NUM) strerror(E_NUM)
      static char *
      strerror(int e_num);
      static char buf[132];
      static char *
      strerror(int e_num)
      {
        sprintf(buf, "Error #%d", e_num);
        return buf;
      }
#endif

#define NAME_BUF_LENGTH MAXFILENAMECHAR

extern OPTS_REC options; 

#ifdef QUASI_ESTIMATOR
void produce_mctlps(void);
void produce_mctlts(FILE *f);
#endif

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */
 /* PRIVATE */
 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

static void write_afro_file(void);
static FILE *open_afro_file(void);

static void write_mos_file(void);
static FILE *open_mos_file(void);

static void write_wasp_file(void);
static FILE *open_wasp_file(void);

static void close_output_file(FILE *fp);

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

void report_end(void)
{
  char buffer[NAME_BUF_LENGTH];
  char line[NAME_BUF_LENGTH];
  unsigned long int tat, tbt;
  
  if (options.use_logging == YES) {
    strcpy(line, "\n            Finished: ");
    get_timing(buffer);
    log_to_file(NULL, strcat(line, buffer));
    sprintf(buffer,
	    "\n    Generated totals: %lu triangles, %lu vertices, %lu edges",
	    (tat = (unsigned long int)total_of_triangles()),
	    (unsigned long int)total_of_vertices(),
	    (unsigned long int)total_of_edges());
    log_to_file(NULL, buffer);
    sprintf(buffer,
	    "         %lu bad-shape triangle(s), i.e. %f percent.",
	    (tbt = (unsigned long int)total_of_bad_triangles()),
	    100*(float)tbt/(float)tat);
    log_to_file(NULL, buffer);
  }
}


 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

void output_mesh()
{
  char est_name[NAME_BUF_LENGTH] = { '\0' };
  FILE *stream;
  
  report_end();
#ifdef QUASI_ESTIMATOR
  if (options.produce_quasi_estimate) {
    strcat(est_name, options.in_file_name);
    if (has_suffix(est_name, IN_SUFFIX))
      est_name[strlen(est_name)-strlen(IN_SUFFIX)] = '\0';
    strcat(est_name, EST_SUFFIX);

    if ((stream = fopen(est_name, "w")) == NULL)
      manip_error("Failed to open out stream for estimate");
    options.mctl_by_func = abs(options.mctl_by_func);
    produce_mctlts(stream);
    fclose(stream);
  }
#endif
  if (!options.no_output) {
    switch (options.output_format) {
    case 3:
      write_wasp_file ();
      break;
    case 2:
      write_mos_file();
      break;
    case 1:
    default:
      write_afro_file();
      break;
    }
  }
}



static FILE *
open_afro_file(void)
{
  FILE *file_p;
  char mode[] = "w+";
  
  errno = 0;
  if ((file_p = fopen(options.out_file_name, mode)) == NULL) {
    fprintf(stderr, "Open for file '%s' in mode '%s' failed: \n   %s\n",
	    options.out_file_name, mode, strerror(errno));
  }
  
  fprintf(file_p, "%s\n", options.out_file_name);
  fprintf(file_p, "%12lu %12lu\n", total_of_vertices(), total_of_triangles());
  
  return file_p;
}



static FILE *
open_mos_file(void)
{
  FILE *file_p;
  char mode[] = "w+";
  
  errno = 0;
  if ((file_p = fopen(options.out_file_name, mode)) == NULL) {
    fprintf(stderr, "Open for file '%s' in mode '%s' failed: \n   %s\n",
	    options.out_file_name, mode, strerror(errno));
  }
  
  fprintf(file_p, " %12lu  %12lu  %12lu\n",
          total_of_vertices(),
          total_of_edges(),
          total_of_triangles());
  
  return file_p;
}

static FILE *
open_wasp_file(void)
{
  FILE *file_p;
  char mode[] = "w+";
  
  errno = 0;
  if ((file_p = fopen(options.out_file_name, mode)) == NULL) {
    fprintf(stderr, "Open for file '%s' in mode '%s' failed: \n   %s\n",
	    options.out_file_name, mode, strerror(errno));
  }
  
  fprintf(file_p, "! WASP (C) 2001 Petr Krysl: %12lu nodes, %12lu triangles\n",
          total_of_vertices(),
          total_of_triangles());
  
  return file_p;
}



static void 
close_output_file(FILE *fp)
{
  if (fclose(fp) == EOF) {
    fprintf(stderr, "Close for file '%s' failed: \n   %s\n",
	    options.out_file_name, strerror(errno));
  }
}




static void 
write_afro_file(void)
{
  FILE *fp;
  
  fp = open_afro_file();
  dump_nodes_in_afro_style(fp);
  dump_triangles_in_afro_style(fp);
  close_output_file(fp);
}



static void 
write_mos_file(void)
{
  FILE *fp;
  
  fp = open_mos_file();
  dump_nodes_in_mos_style(fp);
  dump_edges_in_mos_style(fp);
  dump_triangles_in_mos_style(fp);
  close_output_file(fp);
}

static void 
write_wasp_file(void)
{
  FILE *fp;
  
  fp = open_wasp_file();
  dump_nodes_in_wasp_style(fp);
  dump_triangles_in_wasp_style(fp);
  close_output_file(fp);
}
