/*  
    =========================================================================

    targe2: a program for "T"riangulation of "AR"bitrary "GE"ometries in "2"D

    =========================================================================

                      Copyright: 1993, 1994 Petr Krysl 

       Czech Technical University in Prague, Faculty of Civil Engineering,
           Dept. Structural Mechanics, 166 29 Prague, Czech Republic,
                          e-mail: pk@aix.fsv.cvut.cz

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    =========================================================================
*/

#include "targe2cf.h"
#include "ttypes.h"

#include <string.h>
#include "utils.h"

#include "tops.h"
#include "vertex.h"
#include "curve.h"
#include "loops.h"
#include "subregs.h"
#include "mesh.h"
#include "cells.h"


#ifdef X_XFACE

#include "xiface.h"

#endif

#ifdef DOS_GXFACE

#include "views.h"
#include "scr.h"
#include "status.h"

#endif


SUBREG_LIST subreg_list;

static int mesh_it(NODE ptr, SUBREG_REC *s_r);

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */
 /* make_subregs_list  */

void 
make_subregs_list()
{
  subreg_list = make_list();
}

#ifdef NEED_PRINT
int 
print_loop(NODE ptr, CURVE_REF_REC *c_r)
{
  printf("\nCurve ref: %d ",
	 c_r->curve_sense * get_curve_id(c_r->curve_p));
}

int 
print_int_loops(NODE ptr, C_REFS_LIST *c_r_l)
{
  printf("\nHole:  ");
  for_all_list(c_r_l, NULL, print_loop);
}
#endif

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

BOOLEAN 
add_subreg(int srid,
           int pid, int mid, int sid, float thickness, char *group,
           C_REFS_LIST ext_loop,
           C_REFS_LIST_LIST int_loops)
{
  SUBREG_REC *n;
  make_and_add_to_tail_val(subreg_list, n, SUBREG_REC, "add_subreg", srid); 
  n->subreg_id   = srid;
  n->property_id = pid;
  n->material_id = mid;
  n->subsoil_id  = sid;
  n->thickness   = thickness;
  DUPLICATE_STRING(n->group, group);
  n->ext_loop    = ext_loop;
  n->int_loops   = int_loops;
  n->meshed      = NO;
/*for_all_list(ext_loop, NULL, print_loop);
  for_all_list(int_loops, NULL, print_int_loops); */
  return YES;
}


 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

void 
subreg_data(SUBREG_REC *subreg,
            int *pid, int *mid, int *sid, float *thickness, char **group)
{
  *pid       = subreg->property_id;
  *mid       = subreg->material_id;
  *sid       = subreg->subsoil_id;
  *thickness = subreg->thickness;
  *group     = subreg->group;
}

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

static float bbox_llx = 0, bbox_lly = 0, bbox_urx = 0, bbox_ury = 0;

#define BIGBIGNUMBER 1.e30

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

static int 
update_bbox(NODE ptr, SUBREG_REC *s_r)
{
  scan_loop_extent((C_REFS_LIST)s_r->ext_loop,
                   &bbox_llx, &bbox_lly, &bbox_urx, &bbox_ury);
  return 1;
}


void
set_subregions_bbox(float minx, float miny, float maxx, float maxy)
{
  bbox_llx = minx;
  bbox_lly = miny;
  bbox_urx = maxx;
  bbox_ury = maxy;
}

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

void
subregions_bbox(float *minx, float *miny, float *maxx, float *maxy)
{
  if (bbox_llx == 0 &&   bbox_lly == 0 && bbox_urx == 0 && bbox_ury == 0) {
    bbox_llx =  BIGBIGNUMBER;
    bbox_lly =  BIGBIGNUMBER;
    bbox_urx = -BIGBIGNUMBER;
    bbox_ury = -BIGBIGNUMBER;
    for_all_list(subreg_list, NULL, update_bbox);
  } else {                      /* extent was specified in input */
    /* VOID */
  }
  *minx = bbox_llx;
  *miny = bbox_lly;
  *maxx = bbox_urx;
  *maxy = bbox_ury;
}

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */
 /* mesh_subregions  */

/* kresleni:  spodni dolni roh (tlx, tly) a horni pravy roh (thx, thy)
   Tato cisla jsou aktualizovana pri kazdem kresleni podoblasti.
 */
static float tlx = BIGBIGNUMBER, tly = BIGBIGNUMBER;
static float thx = -BIGBIGNUMBER, thy = -BIGBIGNUMBER;
/* kresleni:  rx, ry je referencni bod uprostred podoblasti
   dx, dy jsou rozmery oblasti (takze napr. pravy horni roh je
   rx + dx/2, ry + dy/2).
 */
static float rx, ry, dx, dy;

#ifdef X_XFACE
static int
fit_and_draw_region(NODE ptr, NODE elem)
{
  WCRec p;
  p.x = rx; p.y = ry; p.z = 0;
  EVSetDimsVC((EView *)elem, dx, dy);
  EVSetVRP((EView *)elem, &p);
  EVForceRedisplay((EView *)elem);
  return 1;
}
#endif


#define NO_VIEW_CHANGE 1

static int
compute_and_compare_bbox(SUBREG_REC *s1, SUBREG_REC *s2)
{
  float lx = BIGBIGNUMBER, ly = BIGBIGNUMBER;
  float hx = -BIGBIGNUMBER, hy = -BIGBIGNUMBER;
  float a1 = 0, a2 = 0;

  if (s1 != NULL) {
    scan_loop_extent((C_REFS_LIST)s1->ext_loop, &lx, &ly, &hx, &hy);
    a1 = (hx - lx) * (hy - ly);
    tlx = min(tlx, lx); tly = min(tly, ly);
    thx = max(thx, hx); thy = max(thy, hy);
  }

  if (s2 != NULL) {
    scan_loop_extent((C_REFS_LIST)s2->ext_loop, &lx, &ly, &hx, &hy);
    a2 = (hx - lx) * (hy - ly);
    tlx = min(tlx, lx); tly = min(tly, ly);
    thx = max(thx, hx); thy = max(thy, hy);
  }
  
  return a1 >= a2;
}


void 
mesh_subregions()
{
#ifdef X_XFACE
# ifdef NO_VIEW_CHANGE
  {
    if (list_length(subreg_list) > 1) {
      sort_list_func(subreg_list, compute_and_compare_bbox);
    } else {
      compute_and_compare_bbox((SUBREG_REC *)get_list_head(subreg_list), NULL);
    }
    rx = (tlx + thx) / 2; ry = (tly + thy) / 2;
    dx = 1.05 * (thx - tlx); dy = 1.05 * (thy - tly);
    if (options.X_xface)
      EMDispatchToDependentViews(ESIModel(), fit_and_draw_region, NULL);
  }
# endif
#endif  
  for_all_list(subreg_list, NULL, mesh_it);
#ifdef X_XFACE
  rx = (tlx + thx) / 2;
  ry = (tly + thy) / 2;
  dx = 1.05 * (thx - tlx);
  dy = 1.05 * (thy - tly);
  if (options.X_xface)
    EMDispatchToDependentViews(ESIModel(), fit_and_draw_region, NULL);
#endif
}



static int 
mesh_it(NODE ptr, SUBREG_REC *s_r)
{
  construct_front();
  construct_cells((C_REFS_LIST)s_r->ext_loop);
  mesh_loop((SUBREG_REC *)s_r, (C_REFS_LIST)s_r->ext_loop);
  if (s_r->int_loops != NULL)
    for_all_list(s_r->int_loops, (SUBREG_REC *)s_r, mesh_loop);
#ifdef X_XFACE
  if (options.X_xface) {
#   ifndef NO_VIEW_CHANGE
    float lx = BIGBIGNUMBER, ly = BIGBIGNUMBER;
    float hx = -BIGBIGNUMBER, hy = -BIGBIGNUMBER;
    scan_loop_extent((C_REFS_LIST)s_r->ext_loop, &lx, &ly, &hx, &hy);
    fprintf(stderr, "%g %g %g %g\n", lx, ly, hx, hy);
    tlx = min(tlx, lx); tly = min(tly, ly);
    thx = max(thx, hx); thy = max(thy, hy);
    rx = (lx + hx) / 2; ry = (ly + hy) / 2;
    dx = 1.05 * (hx - lx); dy = 1.05 * (hy - ly);
    EMDispatchToDependentViews(ESIModel(), fit_and_draw_region, NULL);
#   endif
    xiface_loop(YES, "After meshing loops");
  }
#endif
#ifdef DOS_GXFACE
  if (options.DOS_gxface) {
    /* RAW: message */
    Print_Status(1);
  }
#endif  
  mesh_interior((SUBREG_REC *)s_r);
  destruct_cells();
  destruct_front();
  return 1;
}

