
#include "targe2cf.h"
#include "ttypes.h"
#include "qtree.h"


#include <sys/time.h>

static struct timeval start_up_time; 

int gettimeofday (struct timeval *Tp, struct timezone *Tzp);

long unsigned int 
msec_timer(BOOLEAN start_timing) 
{
  struct timeval stv;

  if (gettimeofday(&stv, (struct timezone *) 0) < 0)
    fprintf(stderr, "Error: failed in gettimeofday \n");
  
  if (start_timing) {
    start_up_time.tv_sec  = stv.tv_sec;
    start_up_time.tv_usec = stv.tv_usec;
    return 0;
  } else 
    return  ((stv.tv_sec - start_up_time.tv_sec - 1) * 1000 +
	     (stv.tv_usec - start_up_time.tv_usec + 1000000) / 1000);
}


static int depth = 20, cycles = 1000000;
static double at_x, at_y;

typedef struct bg_fe_node {
    double x, y;
    double d;
}              bg_fe_node;

typedef struct bg_fe_elem {
    bg_fe_node  *n1, *n2, *n3;
}              bg_fe_elem;

static bg_fe_node n1, n2, n3;
static bg_fe_elem el, *elp;


void
set_subregions_bbox(float minx, float miny, float maxx, float maxy);

static int 
ccw (double x0, double y0, double x1, double y1, double x2, double y2);

static BOOLEAN
is_inside_triangle(double x, double y, /* point */
                   double x1, double y1, /* triangle vertices */
                   double x2, double y2,
                   double x3, double y3)
{
  int orient0, orient1, orient2;

  orient0 = ccw (x1, y1, x2, y2, x, y);
  orient1 = ccw (x2, y2, x3, y3, x, y);
  orient2 = ccw (x3, y3, x1, y1, x, y);

  /* wildcard: go either way */
  if (orient0 == 0)
    orient0 = orient1? orient1: orient2;
  if (orient1 == 0)
    orient1 = orient0;
  if (orient2 == 0)
    orient2 = orient0;

  return (orient0==orient1 && orient1==orient2);
}


/*
	This routine determines if the progression of points p0->p1->p2
is counter-clockwise or not.  The return value is
	 1) counter-clockwise, 
	 0) if the the points are linear, and 
	-1) if they are clockwise.
Another way of interpreting this data is: 
	 1) point 2 is to the left of the line p0->p1, 
	 0) point 2 is on the line, and 
	-1) point 2 is to the right.
*/
static int 
ccw (double x0, double y0, double x1, double y1, double x2, double y2)
{
  double dx1, dx2, dy1, dy2;
  double area1, area2;

  dx1= x1 - x0, 	dy1= y1 - y0;
  dx2= x2 - x0, 	dy2= y2 - y0;

  area1= dx1*dy2;
  area2= dy1*dx2;

  if (area1 > area2)
	return (1);	/* counter-clockwise, left */

  if (area1 < area2)
	return (-1);	/* clockwise, right */

  return (0);	/* linear */
}



float
interpolate_on_triangle(float x, float y,
                        float x1, float y1,
                        float x2, float y2,
                        float x3, float y3,
                        float v1, float v2, float v3);


float
interpolate_on_triangle(float x, float y,
                        float x1, float y1,
                        float x2, float y2,
                        float x3, float y3,
                        float v1, float v2, float v3)
{
    double L1, L2, L3, two_A, a1, a2, a3, b1, b2, b3, c1, c2, c3, two_A_inv;

    a1 = x2 * y3 - x3 * y2;
    a2 = x3 * y1 - x1 * y3;
    a3 = x1 * y2 - x2 * y1;
    two_A = a1 + a2 + a3;
    two_A_inv = 1 / two_A;
    b1 = y2 - y3;
    b2 = y3 - y1;
    b3 = y1 - y2;
    c1 = x3 - x2;
    c2 = x1 - x3;
    c3 = x2 - x1;
    L1 = (a1 + b1 * x + c1 * y) * two_A_inv;
    L2 = (a2 + b2 * x + c2 * y) * two_A_inv;
    L3 = (a3 + b3 * x + c3 * y) * two_A_inv;
    return (L1 * v1 + L2 * v2 + L3 * v3);
}



long unsigned int 
msec_timer(BOOLEAN start_timing);
void
build_quadtree(int depth);
void
test_quadtree_descent(void);



void
build_quadtree(int depth)
{
    int i;
    float X = 0, Y = 0, Dim = 2;

    depth--;
    
    set_subregions_bbox(-1, -1, 1, 1);
    X = Y = 0;
    Dim *= 0.999;
    for (i = 0; i < depth; i++) Dim /= 2;
    qt_extend_qtree(X, Y, Dim, 0, 0);
    for (i = 0; i < depth; i++) qt_extend_qtree(X, Y+i*Dim, Dim, 0, 0);
    for (i = 0; i < depth; i++) qt_extend_qtree(X+i*Dim, Y, Dim, 0, 0);
    for (i = 0; i < depth; i++) qt_extend_qtree(X, Y-i*Dim, Dim, 0, 0);
    for (i = 0; i < depth; i++) qt_extend_qtree(X-i*Dim, Y, Dim, 0, 0);
    at_x = X;
    at_y = Y;
    Dim = qt_leaf_cdim_at(at_x, at_y);
}


void
test_quadtree_descent(void)
{
    int i;
    float Dim = 0;
    
    build_quadtree(depth);
    msec_timer(YES);
    for  (i = 0; i < cycles; i++) {
        Dim = qt_leaf_cdim_at(at_x, at_y);
    }
    fprintf(stderr, "Access of leaf at depth %d for %d cycles: %lu [msec]\n",
            depth, cycles, msec_timer(NO));
}


void
test_triangle_in_test(void)
{
    int i;
    float X = 0, Y = 0, Dim = (double)1/(double)cycles, ad;
    double xmin, ymin, xmax, ymax;
    BOOLEAN is_in;

    
    n1.x = 0; n1.y = 0; n1.d = Dim;
    n2.x = 0; n2.y = 0; n2.d = Dim;
    n3.x = 0; n3.y = 0; n3.d = Dim;
    el.n1 = &n1;
    el.n2 = &n2;
    el.n3 = &n3;
    elp = &el;
    X = Y = 0;
    msec_timer(YES);
    for  (i = 0; i < cycles; i++) {
        is_in = is_inside_triangle(X, Y,
                                   elp->n1->x, elp->n1->y,
                                   elp->n2->x, elp->n2->y,
                                   elp->n3->x, elp->n3->y);
    }
    fprintf(stderr, "In-triangle test for %d cycles: %lu [msec]\n",
            cycles, msec_timer(NO));
    msec_timer(YES);
    for  (i = 0; i < cycles; i++) {
        xmin = min_of_3(elp->n1->x, elp->n2->x, elp->n3->x);
        ymin = min_of_3(elp->n1->y, elp->n2->y, elp->n3->y);
        xmax = max_of_3(elp->n1->x, elp->n2->x, elp->n3->x);
        ymax = max_of_3(elp->n1->y, elp->n2->y, elp->n3->y);
        is_in = (X >= xmin && X <= xmax && Y >= ymin && Y <= ymax);
    }
    fprintf(stderr, "Bounding box test for %d cycles: %lu [msec]\n",
            cycles, msec_timer(NO));
    msec_timer(YES); 
    for  (i = 0; i < cycles; i++) {
        ad = interpolate_on_triangle(X, Y,
                                     elp->n1->x, elp->n1->y,
                                     elp->n2->x, elp->n2->y,
                                     elp->n3->x, elp->n3->y,
                                     elp->n1->d,
                                     elp->n2->d,
                                     elp->n3->d);
    }
    fprintf(stderr, "Interpolation for %d cycles: %lu [msec]\n",
            cycles, msec_timer(NO));
}





extern OPTS_REC options; /* this structure holds all options */

int 
main(int argc, char **argv)
{
    test_triangle_in_test();
    test_quadtree_descent();

    return 1;
}
