
#include "getoptP.h"
#include "dictP.h"

static void
usage(int argc, char **argv);
static void
choke(int argc, char **argv, char *blurb);
static void
process(int argc, char **argv);



int
main(int argc, char **argv)
{
  if (!tokens_open_file("-")) {
    choke(argc, argv, "Cannot read standard input");
  }
  process(argc, argv);
}


static void
choke(int argc, char **argv, char *blurb)
{
  fprintf(stderr, "%s\n", blurb);
  usage(argc, argv);
}


static void
usage(int argc, char **argv)
{
  fprintf(stderr, "%s < inp > outp\n", argv[0]);
  exit(0);
}


static char buf[256];       

#define NEXT_LINE(EXPECTING)                                                \
   {                                                                        \
     line++;                                                                \
     if (!tokens_next_line()) {                                             \
       sprintf(buf, "Failed reading standard input; line %d: expecting %s", \
               EXPECTING);                                                  \
       choke(argc, argv, buf);                                              \
     }                                                                      \
   }

/* Macros for access to the parsed line */
#define TOKEN_DBL(ith, var)                                                  \
  {                                                                          \
    if (tokens_total_of_tokens() >= ith) {                                   \
      if ((tokens_type_of_token(ith) == TOKENS_NUMBER)) {                    \
        var = tokens_token_as_double(ith);                                   \
      } else {                                                               \
        sprintf(buf, "Expecting number in line %d, pos. %d.  Got %s",        \
                line, ith, tokens_token_as_string(ith));                     \
        choke(argc, argv, buf);                                              \
      }                                                                      \
    } else {                                                                 \
      sprintf(buf,                                                           \
              "Expecting token in line %d, pos. %d.  Got only %d tokens",    \
              line, ith, tokens_total_of_tokens());                          \
      choke(argc, argv, buf);                                                \
    }                                                                        \
  }
#define TOKEN_INT(ith, var)                                                  \
  {                                                                          \
    if (tokens_total_of_tokens() >= ith) {                                   \
      if ((tokens_type_of_token(ith) == TOKENS_NUMBER)) {                    \
        var = tokens_token_as_int(ith);                                      \
      } else {                                                               \
        sprintf(buf, "Expecting number in line %d, pos. %d.  Got %s",        \
                line, ith, tokens_token_as_string(ith));                     \
        choke(argc, argv, buf);                                              \
      }                                                                      \
    } else {                                                                 \
      sprintf(buf,                                                           \
              "Expecting token in line %d, pos. %d.  Got only %d tokens",    \
              line, ith, tokens_total_of_tokens());                          \
      choke(argc, argv, buf);                                                \
    }                                                                        \
  }

#define NEW(TYPE) ((TYPE *) malloc(sizeof(TYPE)))

#define CHECK_NEW(ptr)                          \
   {                                            \
     if (ptr == NULL) {                         \
       choke(argc, argv, "Failed malloc");      \
     }                                          \
   }

static void
read_vertices(int argc, char **argv);
static void
read_edges(int argc, char **argv);
static void
read_triangles(int argc, char **argv);
static void
add_midside_vertices(int argc, char **argv);
static void
write_header(int argc, char **argv);
static void
write_vertices(int argc, char **argv);
static void
write_triangles(int argc, char **argv);

static int line = 0;

static int nvertices, nedges, ntriangles;
static struct dict_dictionary *vertices = NULL;
static struct dict_dictionary *edges = NULL;
static struct dict_dictionary *triangles = NULL;


static void
process(int argc, char **argv)
{
  NEXT_LINE("# vertices   # edges   # triangles");
  TOKEN_INT(1, nvertices);
  TOKEN_INT(2, nedges);
  TOKEN_INT(3, ntriangles);
  read_vertices(argc, argv);
  read_edges(argc, argv);
  read_triangles(argc, argv);
  add_midside_vertices(argc, argv);
  write_header(argc, argv);
  write_vertices(argc, argv);
  write_triangles(argc, argv);
}

#define FOR(_Counter_, _Up_)                                      \
   for (_Counter_ = 0; _Counter_ < _Up_; _Counter_++)

typedef struct vertex_t {
  int id;
  double x, y;
  int is_midside;
}              vertex_t;

typedef struct edge_t {
  int id;
  int v1, v2;
  int midv;
  int curve;
}              edge_t;

typedef struct triangle_t {
  int id;
  int e1, e2, e3;
  int subreg;
}              triangle_t;


static void
read_vertices(int argc, char **argv)
{
  int i;
  int id, curve;
  double x, y;
  vertex_t *v;
  
  vertices = dict_create_dictionary();
  if (vertices == NULL)
    choke(argc, argv, "Failed to create dictionary");

  FOR(i, nvertices) {
    NEXT_LINE("vertex id     x     y     curve id");
    TOKEN_INT(1, id);
    TOKEN_DBL(2, x);
    TOKEN_DBL(3, y);
    TOKEN_INT(4, curve);
    v = NEW(vertex_t);
    CHECK_NEW(v);
    v->id         = id;
    v->x          = x;
    v->y          = y;
    v->is_midside = 0;
    dict_add_obj_int_name(vertices, v, v->id);
  }
}


static void
read_edges(int argc, char **argv)
{
  int i;
  int id, curve;
  int v1, v2;
  edge_t *e;
  
  edges = dict_create_dictionary();
  if (edges == NULL)
    choke(argc, argv, "Failed to create dictionary");

  FOR(i, nedges) {
    NEXT_LINE("edge id     v1   v2      curve id");
    TOKEN_INT(1, id);
    TOKEN_INT(2, v1);
    TOKEN_INT(3, v2);
    TOKEN_INT(4, curve);
    e = NEW(edge_t);
    CHECK_NEW(e);
    e->id    = id;
    e->v1    = v1;
    e->v2    = v2;
    e->midv  = -1;
    e->curve = curve;
    dict_add_obj_int_name(edges, e, e->id);
  }
}



static void
read_triangles(int argc, char **argv)
{
  int i;
  int id, subreg;
  int e1, e2, e3;
  triangle_t *t;
  
  triangles = dict_create_dictionary();
  if (triangles == NULL)
    choke(argc, argv, "Failed to create dictionary");

  FOR(i, ntriangles) {
    NEXT_LINE("triangle id     e1   e2   e3     subreg id");
    TOKEN_INT(1, id);
    TOKEN_INT(2, e1);
    TOKEN_INT(3, e2);
    TOKEN_INT(4, e3);
    TOKEN_INT(5, subreg);
    t = NEW(triangle_t);
    CHECK_NEW(t);
    t->id     = id;
    t->e1     = e1;
    t->e2     = e2;
    t->e3     = e3;
    t->subreg = subreg;
    dict_add_obj_int_name(triangles, t, t->id);
  }
}


static void
gen_for_edge(int argc, char **argv, edge_t *e);


static void
add_midside_vertices(int argc, char **argv)
{
  triangle_t *t;
  edge_t *e;
  vertex_t *v;

  t = NULL;
  while ((t = (triangle_t *)dict_next_obj(triangles, t)) != NULL) {
    e = (edge_t *)dict_find_obj_with_int_name(edges, t->e1);
    gen_for_edge(argc, argv, e);
    e = (edge_t *)dict_find_obj_with_int_name(edges, t->e2);
    gen_for_edge(argc, argv, e);
    e = (edge_t *)dict_find_obj_with_int_name(edges, t->e3);
    gen_for_edge(argc, argv, e);
  }
}


static void
gen_for_edge(int argc, char **argv, edge_t *e)
{
  vertex_t *v = NULL;
  vertex_t *v1, *v2;
  
  if (e == NULL) {
    choke(argc, argv, "NULL edge");
  } else {
    if (e->midv < 0) {
      v = NEW(vertex_t); CHECK_NEW(v);
      v->id = ++nvertices;
      v1 = (vertex_t *)dict_find_obj_with_int_name(vertices, e->v1);
      v2 = (vertex_t *)dict_find_obj_with_int_name(vertices, e->v2);
      if (v1 == NULL || v2 == NULL) {
        choke(argc, argv, "NULL vertex");
      } 
      v->x  = 0.5 * (v1->x + v2->x);
      v->y  = 0.5 * (v1->y + v2->y);
      v->is_midside = 1;
      dict_add_obj_int_name(vertices, v, v->id);
      e->midv = v->id;
    }
  }
}


static void
write_vertices(int argc, char **argv)
{
  vertex_t *v = NULL;
  
  while ((v = (vertex_t *)dict_next_obj(vertices, v)) != NULL) {
    fprintf(stdout, "%d \t%g \t%g\n", v->id, v->x, v->y);
  }
}


static void
find_corners(int argc, char **argv,
             edge_t *e1, edge_t *e2, edge_t *e3,
             vertex_t **v1, vertex_t **v2, vertex_t **v3);


static void
write_triangles(int argc, char **argv)
{
  triangle_t *t;
  edge_t *e1, *e2, *e3;
  vertex_t *v1, *v2, *v3;
  vertex_t *v4, *v5, *v6;

  t = NULL;
  while ((t = (triangle_t *)dict_next_obj(triangles, t)) != NULL) {
    e1 = (edge_t *)dict_find_obj_with_int_name(edges, t->e1);
    e2 = (edge_t *)dict_find_obj_with_int_name(edges, t->e2);
    e3 = (edge_t *)dict_find_obj_with_int_name(edges, t->e3);
    find_corners(argc, argv, e1, e2, e3, &v1, &v2, &v3);
    v4 = (vertex_t *)dict_find_obj_with_int_name(vertices, e1->midv);
    v5 = (vertex_t *)dict_find_obj_with_int_name(vertices, e2->midv);
    v6 = (vertex_t *)dict_find_obj_with_int_name(vertices, e3->midv);
    fprintf(stdout, "%d \t%d \t%d \t%d \t%d \t%d \t%d\n",
            t->id, v1->id, v2->id, v3->id, v4->id, v5->id, v6->id);
  }
}


static vertex_t *
common_vertex(edge_t *e1, edge_t *e2);

static void
find_corners(int argc, char **argv,
             edge_t *e1, edge_t *e2, edge_t *e3,
             vertex_t **v1, vertex_t **v2, vertex_t **v3)
{
  *v1 = common_vertex(e3, e1);
  *v2 = common_vertex(e1, e2);
  *v3 = common_vertex(e2, e3);
  
  if (*v1 == NULL || *v2 == NULL || *v3 == NULL) {
    choke(argc, argv, "No corner vertex");
  }
}


static vertex_t *
common_vertex(edge_t *e1, edge_t *e2)
{
  vertex_t *vs1e, *ve1e;
  vertex_t *vs2e, *ve2e;

  vs1e = (vertex_t *)dict_find_obj_with_int_name(vertices, e1->v1);
  ve1e = (vertex_t *)dict_find_obj_with_int_name(vertices, e1->v2);
  
  vs2e = (vertex_t *)dict_find_obj_with_int_name(vertices, e2->v1);
  ve2e = (vertex_t *)dict_find_obj_with_int_name(vertices, e2->v2);

  if      (vs1e == vs2e) return vs2e;
  else if (vs1e == ve2e) return ve2e;
  else if (ve1e == vs2e) return vs2e;
  else if (ve1e == ve2e) return ve2e;
  else                   return NULL;
}


static void
write_header(int argc, char **argv)
{
  fprintf(stdout, "# %s (C)1996 P. Krysl\n", argv[0]);
  fprintf(stdout, "%d %d\n", nvertices, ntriangles);
}

