/*  
    =========================================================================

    targe2: a program for "T"riangulation of "AR"bitrary "GE"ometries in "2"D

    =========================================================================

                      Copyright: 1993, 1994 Petr Krysl 

       Czech Technical University in Prague, Faculty of Civil Engineering,
           Dept. Structural Mechanics, 166 29 Prague, Czech Republic,
                          e-mail: pk@aix.fsv.cvut.cz

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    =========================================================================
*/

#include "targe2cf.h"
#include "ttypes.h"

#include "utils.h"

#include "tops.h"
#define VERTEX_PRIVATE_HDR
#include "vertex.h"
#include "edges.h"
#include "curve.h"
#include "mctlp.h"
#include "mesh.h"

#ifdef X_XFACE
#include <stdio.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xresource.h>
#include <X11/Intrinsic.h>
#define HAVE_BOOLEAN
#include "xiface.h"
#endif

extern OPTS_REC options; /* import global options */

VERTEX_LIST vertex_list;

static CURVE_REC special_curve = {
  { NULL, DISCONTINUOUS_SPECIAL_CURVE, NO, NULL, NULL, AS_DEFINED_SENSE },
  { 0 },
}; 

static long vert_total = 0;

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

void 
make_vertex_list()
{
  vertex_list = MAKE_LIST();
}

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

long int 
total_of_vertices()
{
  return LIST_LENGTH(vertex_list);
}

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

VERTEX_REC *
vertex_near(float fx, float fy)
{
  float tol = options.merge_tolerance;
  VERTEX_REC *vp;

  vp = (VERTEX_REC *)GET_LIST_NEXT(vertex_list, NULL);
  while (vp != NULL) {
    if (fabs((double)(fx - vp->coords.x)) < tol)
      if (fabs((double)(fy - vp->coords.y)) < tol)
	return vp;
    vp = (VERTEX_REC *)GET_LIST_NEXT(vertex_list, vp);
  }
  return NULL;
}

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

VERTEX_REC *
add_vertex(int id, float x, float y,
                       float cdim, float mf, float ang)
{
  VERTEX_REC *v_p;
  long int i;

  vert_total++;
  if (id != 0)
    i = id;
  else
    i = vert_total;
  MAKE_AND_ADD_TO_TAIL_VAL(vertex_list, v_p, VERTEX_REC, "add_vertex", i);
  v_p->id           = vert_total; /* unique identifiers! */
  v_p->coords.x     = x;
  v_p->coords.y     = y;
  v_p->connectivity = 0;
  for (i = 0; i < MAX_CONNECTIVITY; i++)
    v_p->used_by_edges[i] = NULL;
  v_p->parent_curve = NULL;
  if (cdim <= 0.0) {
    v_p->stored_cdim  = get_mesh_cdim(x, y);
#ifdef ANISOTROPIC_MESH
    v_p->stored_mf    = get_mesh_2nd_dir_magn_factor(x, y);
    v_p->stored_ang   = get_mesh_1st_dir_angle(x, y);
#endif    
  } else {
    v_p->stored_cdim  = cdim;
#ifdef ANISOTROPIC_MESH
    v_p->stored_mf    = mf;
    v_p->stored_ang   = ang;
#endif
  }
  return v_p;
}

void
destroy_vertex(VERTEX_REC *v_p)
{
  DELETE_LIST_NODE(vertex_list, v_p);
}


 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

#ifndef vertex_data
void 
vertex_data(VERTEX_REC *v, int *id, float *x, float *y)
{
  *id = v->id;
  *x  = v->coords.x;
  *y  = v->coords.y;
}
#endif

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

#ifndef set_vertex_curve_p
void 
set_vertex_curve_p(VERTEX_REC *v_p, CURVE_REC *c_p) 
{
  v_p->parent_curve = c_p; 
}
#endif

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

#ifndef get_stored_cdim
float 
get_stored_cdim(VERTEX_REC *v_p)
{
  return v_p->stored_cdim;
}
#endif


#ifndef get_stored_mf
float 
get_stored_mf(VERTEX_REC *v_p)
{
  return v_p->stored_mf;
}
#endif


#ifndef get_stored_angle
float 
get_stored_angle(VERTEX_REC *v_p)
{
  return v_p->stored_ang;
}
#endif

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

#ifndef get_vertex_coords
void 
get_vertex_coords(VERTEX_REC *v_p, float *x, float *y)
{
  *x = v_p->coords.x;
  *y = v_p->coords.y;
}
#endif

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

#ifndef set_vertex_coords
void 
set_vertex_coords(VERTEX_REC *v_p, float x, float y)
{
  v_p->coords.x = x;
  v_p->coords.y = y;
}
#endif


#ifndef get_vertex_connectivity
int
get_vertex_connectivity(VERTEX_REC *v_p)
{
  return v_p->connectivity;
}
#endif

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

int 
add_vert_reference(VERTEX_REC *v_p, EDGE_REC *e_p)
{
  int i;
  
  for (i = 0; i < MAX_CONNECTIVITY; i++) {
    if (v_p->used_by_edges[i] == e_p)
      return 1;
    else
      if (v_p->used_by_edges[i] == NULL) {
	v_p->used_by_edges[i] = e_p;
        v_p->connectivity++;
	return 1;
      }
  }
  manip_error("Reference array overflow");
  return 0;
}

int 
remove_vert_reference(VERTEX_REC *v_p, EDGE_REC *e_p)
{
  int i, k;
  
  for (i = 0; i < MAX_CONNECTIVITY; i++) {
    if (v_p->used_by_edges[i] == e_p) {
      for (k = i; k < MAX_CONNECTIVITY-1; k++) {
        v_p->used_by_edges[k] = v_p->used_by_edges[k+1];
      }
      v_p->connectivity--;
      return 1;
    }
  }
  return 0;
}

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

int 
get_vert_references(VERTEX_REC *v_p, int *n, EDGE_REC **ea_p)
{
  int i;
  
  (*n) = 0;
  for (i = 0; i < MAX_CONNECTIVITY; i++) {
    if (v_p->used_by_edges[i] != NULL) 
      ea_p[(*n)++] = v_p->used_by_edges[i];
    else
      return 1;
  }
  manip_error("Too numerous connectivity.");
  return 0;
}

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

static void
smooth_laplacian(void);
static void
smooth_anisotropic(void);


void 
smooth_pass(int smoother_type)
{
  switch (smoother_type) {
  case ANISOTROPIC_SMOOTHING:
    smooth_anisotropic();
    break;
  case LAPLACIAN_SMOOTHING:
  default:
    smooth_laplacian();
    break;
  }
}



#define VCONN get_vertex_connectivity 
#define BETA_F 2
#define VERTEX_CONN_WEIGHT(the_CURR_V_P)                                   \
   (((the_CURR_V_P)->parent_curve != NULL)                                 \
    ? (max(1, (BETA_F*(VCONN(curr_v_p) < 6                                 \
                       ? (2*VCONN(curr_v_p)-8)                             \
                       : (VCONN(curr_v_p) == 6                             \
                          ? 0                                              \
                          : (VCONN(curr_v_p)-4)))                          \
                       + 6)))                                              \
    : (max(1, (BETA_F*((VCONN(curr_v_p)) - 6) + 6)))                       \
    )

static void
smooth_laplacian()
{
  VERTEX_REC_P curr_v_p;
  EDGE_REC_P ep;
  int i;
  float w, wsum, vx, vy;
  VERTEX_REC *v1_p = NULL, *v2_p = NULL, *v_p = NULL;

  v_p = (VERTEX_REC *)GET_LIST_NEXT(vertex_list, NULL);
  while (v_p != NULL) { /* loop over all vertices */
    if (v_p->parent_curve == NULL) {
      vx = 0.; vy = 0.;
      i = 0;
      wsum = 0;
      while ((ep = v_p->used_by_edges[i]) != NULL) {
	get_edge_vertex_p(ep, &v1_p, &v2_p);
	if (v1_p != v_p)
	  curr_v_p = v1_p;
	else
	  curr_v_p = v2_p;
        w = VERTEX_CONN_WEIGHT(curr_v_p);
	vx += w * curr_v_p->coords.x; vy += w * curr_v_p->coords.y;
        wsum += w;
	i++;
      }
      v_p->coords.x = vx / wsum; v_p->coords.y = vy / wsum;
    }
    v_p = (VERTEX_REC *)GET_LIST_NEXT(vertex_list, v_p); 
  }                      /* loop over all vertices */
}



static void
smooth_anisotropic()
{
  VERTEX_REC_P curr_v_p;
  EDGE_REC_P ep;
  int i;
  float w, wsum, vx, vy, L;
  VERTEX_REC *v1_p = NULL, *v2_p = NULL, *v_p = NULL;

  v_p = (VERTEX_REC *)GET_LIST_NEXT(vertex_list, NULL);
  while (v_p != NULL) { /* loop over all vertices */
    if (v_p->parent_curve == NULL) {
      vx = 0.; vy = 0.;
      i = 0;
      wsum = 0;
      while ((ep = v_p->used_by_edges[i]) != NULL) {
	get_edge_vertex_p(ep, &v1_p, &v2_p);
	if (v1_p != v_p)
	  curr_v_p = v1_p;
	else
	  curr_v_p = v2_p;
        L = ideal_edge_length(curr_v_p->coords.x, curr_v_p->coords.y,
                              v_p->coords.x, v_p->coords.y,
                              curr_v_p->stored_cdim,
#ifdef ANISOTROPIC_MESH
                              curr_v_p->stored_mf,
                              curr_v_p->stored_ang
#else
                              1, 0
#endif
                              );
        w = VERTEX_CONN_WEIGHT(curr_v_p);
        w /= L;
	vx += w * curr_v_p->coords.x; vy += w * curr_v_p->coords.y;
        wsum += w;
	i++;
      }
      v_p->coords.x = vx / wsum; v_p->coords.y = vy / wsum;
    }
    v_p = (VERTEX_REC *)GET_LIST_NEXT(vertex_list, v_p); 
  }                      /* loop over all vertices */
}

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

static int fixed_p_total = 0;
static LIST fixed_points = NULL;
static VERTEX_REC *curr_sel_fixed_p = NULL;
static LIST fixed_points_on_curve = NULL;
static LIST tmp_fixed_point_list = NULL;

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

int fixed_vertices_on_curve(CURVE_REC *c)
{
  VERTEX_REC *v_p;
  int num_of_points;

  if (fixed_p_total <= 0) {
    num_of_points = 0;
    goto exit;
  }
  
  if (fixed_points_on_curve) {
    if (!list_empty(fixed_points_on_curve))
      manip_error("Fixed-point list non-empty");
    flush_list(fixed_points_on_curve, NOT_NODES);
  } else
    fixed_points_on_curve = make_list();

  if (!tmp_fixed_point_list)
    tmp_fixed_point_list = make_list();
  
  num_of_points = 0;
  v_p = (VERTEX_REC *)get_list_next(fixed_points, NULL);
  while (v_p != NULL) {
    if (v_p->parent_curve == c) {
      add_to_tail(fixed_points_on_curve, get_from_current(fixed_points));
      num_of_points++;
      v_p = (VERTEX_REC *)get_list_next(fixed_points, NULL);
    } else {
      v_p = (VERTEX_REC *)get_list_next(fixed_points, v_p);
    }
  }
  exit:
    return num_of_points;
}

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

#define COMPUTE_DISTANCES \
 dist_from_prev = line_length(v_p->coords.x, v_p->coords.y, prevx, prevy); \
 dist_from_curr = line_length(v_p->coords.x, v_p->coords.y, *currx, *curry); \
 dist_q1        = line_length(v_p->coords.x, v_p->coords.y, q1x, q1y); \
 dist_h         = line_length(v_p->coords.x, v_p->coords.y, hx, hy); \
 dist_q3        = line_length(v_p->coords.x, v_p->coords.y, q3x, q3y); \
 curr_dist = dist_from_prev

#define DIST_CRITERION (curr_dist < trial_dist && (dist_q1 <trial_radius||dist_q3 <trial_radius||dist_h <trial_radius|| dist_from_curr <2*trial_radius))

VERTEX_REC *sel_fixed_on_curve(float prevx, float prevy,
			       float *currx, float *curry)
{
  VERTEX_REC *v_p;
  VERTEX_REC *sel_v_p = NULL;
  float dist_from_prev, dist_from_curr, selx, sely;
  float trial_dist, curr_dist, max_dist;
  float trial_radius;
  float q1x, q1y, hx, hy, q3x, q3y, dist_q1, dist_h, dist_q3;

  hx           = 0.5*(*currx + prevx); hy  = 0.5*(*curry + prevy);
  q1x          = 0.5*(hx + prevx);    q1y = 0.5*(hy + prevy);
  q3x          = 0.5*(hx + *currx);    q3y = 0.5*(hy + *curry);
  trial_radius = line_length(*currx, *curry, prevx, prevy)/4;
  max_dist     = 6*trial_radius;
  trial_dist   = FLT_MAX;
  v_p = (VERTEX_REC *)get_from_head(fixed_points_on_curve);
  while (v_p != NULL) {
    dist_from_prev = line_length(v_p->coords.x, v_p->coords.y, prevx, prevy); 
    dist_from_curr = line_length(v_p->coords.x, v_p->coords.y, *currx, *curry);
    dist_q1        = line_length(v_p->coords.x, v_p->coords.y, q1x, q1y); 
    dist_h         = line_length(v_p->coords.x, v_p->coords.y, hx, hy); 
    dist_q3        = line_length(v_p->coords.x, v_p->coords.y, q3x, q3y); 
    curr_dist      = dist_from_prev;
    if (curr_dist < trial_dist &&
	(dist_q1 <trial_radius||dist_q3 <trial_radius||
	 dist_h <trial_radius|| dist_from_curr <2*trial_radius)) {
      trial_dist = curr_dist;
      sel_v_p = v_p;
      get_vertex_coords(sel_v_p, &selx, &sely);
      add_to_head(tmp_fixed_point_list, sel_v_p);
    } else
      add_to_tail(tmp_fixed_point_list, v_p);
    v_p = (VERTEX_REC *)get_from_head(fixed_points_on_curve);
  }
  if (sel_v_p) {
    sel_v_p = (VERTEX_REC *)get_from_head(tmp_fixed_point_list);
    *currx = selx;
    *curry = sely;
  }
  while (!list_empty(tmp_fixed_point_list))
    add_to_tail(fixed_points_on_curve, get_from_head(tmp_fixed_point_list));

  if (sel_v_p != NULL) {
    sel_v_p->stored_cdim
      = get_mesh_cdim(sel_v_p->coords.x, sel_v_p->coords.y);
  }
  return sel_v_p;
}

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */



static float
twice_tri_area(float ax, float ay, float bx, float by, float cx, float cy)
/* Returns twice the area of the oriented triangle (a, b, c), i.e., the
   area is positive if the triangle is oriented counterclockwise. */
{
  return (bx - ax)*(cy - ay) - (by - ay)*(cx - ax);
}


static int 
in_circle(float ax, float ay, float bx, float by, float cx, float cy,
          float dx, float dy)
/* Returns TRUE if the point d is inside the circle defined by the
   points a, b, c. See Guibas and Stolfi (1985) p.107. */
{
  return ((ax*ax + ay*ay) * twice_tri_area(bx, by, cx, cy, dx, dy) -
          (bx*bx + by*by) * twice_tri_area(ax, ay, cx, cy, dx, dy) +
          (cx*cx + cy*cy) * twice_tri_area(ax, ay, bx, by, dx, dy) -
          (dx*dx + dy*dy) * twice_tri_area(ax, ay, bx, by, cx, cy) > 0);
}

static VERTEX_REC *
is_delaunay(float fx, float fy, float lx, float ly, VERTEX_REC *s);


void 
select_fixed_point(float ipx, float ipy,
                   float fx, float fy, float lx, float ly,
                   float ecx, float ecy, float rad,
                   VERTEX_REC **selvertexp,
                   float *selvx, float *selvy)
{
  VERTEX_REC *v_p = NULL, *prev_v_p = NULL;
  
  *selvertexp = NULL;
 
#define REDUCTION 0.8

  if (fixed_p_total > 0) {
    v_p = is_delaunay(fx, fy, lx, ly, NULL);
    if (v_p == NULL)
      goto exit;
    prev_v_p = v_p;
    while ((v_p = is_delaunay(fx, fy, lx, ly, v_p)) != prev_v_p) {
      prev_v_p = v_p;
    }
  }
exit:
  if (v_p != NULL) {
    if (   line_length(v_p->coords.x, v_p->coords.y, ecx, ecy)
        <= 2 * line_length(ipx, ipy, ecx, ecy)) {
      *selvertexp = v_p;
      *selvx      = v_p->coords.x;
      *selvy      = v_p->coords.y;
      v_p->stored_cdim  = get_mesh_cdim(*selvx, *selvy);
    }
  }
}


static VERTEX_REC *
is_delaunay(float fx, float fy, float lx, float ly, VERTEX_REC *s)
{
  VERTEX_REC *a;
  float sx = (s != NULL ? s->coords.x : 0);
  float sy = (s != NULL ? s->coords.y : 0);
#define CCW_TRI(fx, fy, lx, ly, sx, sy) \
  (twice_tri_area(fx, fy, lx, ly, sx, sy) > 0)

  a = NULL;
  while ((a = (VERTEX_REC *)get_list_next(fixed_points, a)) != NULL) {
    if (a != s) {
      if (CCW_TRI(fx, fy, lx, ly, a->coords.x, a->coords.y)) {
        if (s == NULL ||
	    in_circle(fx, fy, lx, ly, sx, sy, a->coords.x, a->coords.y)) {
          return a;
        }
      }
    }
  }
  return s; /* the input vertex is delaunay */
}


 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

void adj_fixed_points(VERTEX_REC *selvertexp) 
{
  if (curr_sel_fixed_p != NULL) {
    if (selvertexp == curr_sel_fixed_p) {
      delete_list_node(fixed_points, curr_sel_fixed_p);
      fixed_p_total--;
    }
  }
  curr_sel_fixed_p = NULL;
}

 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

void add_fixed_point(VERTEX_REC *p, int curve_id) 
{
  if (p != NULL) {
    fixed_p_total++;
    if (fixed_points == NULL)
      fixed_points = make_list(); 
    add_to_tail(fixed_points, p);
    if (curve_id == 0)
      p->parent_curve = &special_curve;
    else
      p->parent_curve = find_curve_by_id(curve_id);
#ifdef  X_XFACE 
    if (options.X_xface) {
      WCRec pnts[1];
      GraphicObj *g_p;
      pnts[0].x = p->coords.x;
      pnts[0].y = p->coords.y;
      pnts[0].z = 0;
      g_p = CreateMarker3D(pnts);
      ESIHandleCmd("color magenta");
      ESIHandleCmd("msize 6");
      ESIHandleCmd("mtype 1");
      EGWithMaskChangeAttributes(MSIZE_MASK | MTYPE_MASK | COLOR_MASK, g_p);
      EMAddGraphicsToModel(ESIModel (), g_p);
    }
#endif
  }
}

#ifdef QUASI_ESTIMATOR
 
 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

float mctl_func(float, float);

void produce_mctlps(void)
{
  int i;
  VERTEX_REC *v_p;

  v_p = (VERTEX_REC *)GET_LIST_NEXT(vertex_list, NULL);
  i = 0;
  while (v_p != NULL) {
    printf("m-ctl-point %d xy %f %f near %f influence 1.0\n",
	   ++i, v_p->coords.x, v_p->coords.y,
	   mctl_func(v_p->coords.x, v_p->coords.y));
    v_p = (VERTEX_REC *)GET_LIST_NEXT(vertex_list, v_p); 
  }
}

#endif
 
 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

void       
dump_nodes_in_afro_style(FILE *fp)
{
  VERTEX_REC *vp;
  long id = 0;

  fprintf(fp, "*NODES\n");
  
  vp = (VERTEX_REC *)GET_LIST_NEXT(vertex_list, NULL);
  while (vp != NULL) {
      vp->id = ++id;
    fprintf(fp, "%12lu %12.6f %12.6f\n", vp->id, vp->coords.x, vp->coords.y);
    vp = (VERTEX_REC *)GET_LIST_NEXT(vertex_list, vp);
  }
}



void       
dump_nodes_in_mos_style(FILE *fp)
{
  VERTEX_REC *vp;
  long int id = 0;

  vp = (VERTEX_REC *)GET_LIST_NEXT(vertex_list, NULL);
  while (vp != NULL) {
    /* dumps: id x y parent_curve */
    /* NOTE:  The vertex id is changed here, to account for possible */
    /* deletions during quality enhancments */
    id++;
    if (vp->parent_curve == NULL) {
      fprintf(fp, "%12lu %12.6f %12.6f %12d\n",
              vp->id = id, vp->coords.x, vp->coords.y, 0);
    } else {
      fprintf(fp, "%12lu %12.6f %12.6f %12d\n",
              vp->id = id, vp->coords.x, vp->coords.y,
              get_curve_id(vp->parent_curve));
    }
    vp = (VERTEX_REC *)GET_LIST_NEXT(vertex_list, vp);
  }
}

void       
dump_nodes_in_wasp_style(FILE *fp)
{
  VERTEX_REC *vp;
  long int id = 0;

  fprintf (fp, "NODES\n");
  vp = (VERTEX_REC *)GET_LIST_NEXT(vertex_list, NULL);
  while (vp != NULL) {
    /* dumps: id x y parent_curve */
    /* NOTE:  The vertex id is changed here, to account for possible */
    /* deletions during quality enhancments */
    id++;
    if (vp->parent_curve == NULL) {
      fprintf(fp, "%12lu %g %g 0.0\n",
              vp->id = id, vp->coords.x, vp->coords.y);
    } else {
      fprintf(fp, "%12lu %g %g 0.0 ! on curve %lu\n",
              vp->id = id, vp->coords.x, vp->coords.y, 
              get_curve_id(vp->parent_curve));
    }
    vp = (VERTEX_REC *)GET_LIST_NEXT(vertex_list, vp);
  }
  fprintf (fp, "&\n");
}
